"""
Routines for reading headers and data blocks from Gadget and Subfind snapshot files (format 2 only)
usage e.g.:

import readsnap as rs
header = rs.snapshot_header("snap_063.0") # reads snapshot header
print header.massarr
mass = rs.read_block("snap_063","MASS") 
print mass[0:10]
To print the INFO block just execute the following:
rs.read_block("snap_063","INFO") 


for mutliple file snapshots give e.g. the filename "snap_063" rather than "snap_063.0" to read_block
the returned data block is ordered by particle species even when read from a multiple file snapshot


NS: this script is not (yet) able to read the snap that rely upon the 'massarr' array

"""
import numpy as np
import os
import sys
import math
  
# ----- class for snapshot header ----- 

class snapshot_header:
  def __init__(self, filename):

    if os.path.exists(filename):
      curfilename = filename
    elif os.path.exists(filename+".0"):
      curfilename = filename+".0"
    else:
      print("file not found:", filename)
      sys.exit()
      
    self.filename = filename  
    f = open(curfilename,'rb')    
    blocksize = np.fromfile(f,dtype=np.int32,count=1)
    if blocksize[0] == 8:
      swap = 0
      format = 2
    elif blocksize[0] == 256:
      swap = 0
      format = 1  
    else:
      blocksize.byteswap(True)
      if blocksize[0] == 8:
        swap = 1
        format = 2
      elif blocksize[0] == 256:
        swap = 1
        format = 1
      else:
        print("incorrect file format encountered when reading header of", filename)
        sys.exit()
    
    self.format = format
    self.swap = swap
    
    if format==2:
      f.seek(16, os.SEEK_CUR)
    
    self.npart = np.fromfile(f,dtype=np.int32,count=6)
    self.massarr = np.fromfile(f,dtype=np.float64,count=6)
    self.time = (np.fromfile(f,dtype=np.float64,count=1))[0]
    self.redshift = (np.fromfile(f,dtype=np.float64,count=1))[0]
    self.sfr = (np.fromfile(f,dtype=np.int32,count=1))[0]
    self.feedback = (np.fromfile(f,dtype=np.int32,count=1))[0]
    self.nall = np.fromfile(f,dtype=np.int32,count=6)
    self.cooling = (np.fromfile(f,dtype=np.int32,count=1))[0]
    self.filenum = (np.fromfile(f,dtype=np.int32,count=1))[0]
    self.boxsize = (np.fromfile(f,dtype=np.float64,count=1))[0]
    self.omega_m = (np.fromfile(f,dtype=np.float64,count=1))[0]
    self.omega_l = (np.fromfile(f,dtype=np.float64,count=1))[0]
    self.hubble = (np.fromfile(f,dtype=np.float64,count=1))[0]
    
    if swap:
      self.npart.byteswap(True)
      self.massarr.byteswap(True)
      self.time = self.time.byteswap()
      self.redshift = self.redshift.byteswap()
      self.sfr = self.sfr.byteswap()
      self.feedback = self.feedback.byteswap()
      self.nall.byteswap(True)
      self.cooling = self.cooling.byteswap()
      self.filenum = self.filenum.byteswap()
      self.boxsize = self.boxsize.byteswap()
      self.omega_m = self.omega_m.byteswap()
      self.omega_l = self.omega_l.byteswap()
      self.hubble = self.hubble.byteswap()
     
    f.close()
 






# ----- find offset and size of data block ----- 

def find_block(filename, swap, block, block_num, only_list_blocks=False):
  if (not os.path.exists(filename)):
      print("file not found:", filename)
      sys.exit()

  f = open(filename,'rb')
  f.seek(0, os.SEEK_END)
  filesize = f.tell()
  f.seek(0, os.SEEK_SET)
  
  found = False
  curblock_num = 1
  while ((not found) and (f.tell()<filesize)):
    f.seek(4, os.SEEK_CUR)
    curblock = f.read(4)
    if (block == curblock):
      found = True
    f.seek(8, os.SEEK_CUR)  
        
    curblocksize = (np.fromfile(f,dtype=np.uint32,count=1))[0]
    if swap:
      curblocksize = curblocksize.byteswap()
    
    # - print some debug info about found data blocks -
    #print curblock, curblock_num, curblocksize
    only_list_blocks = False
    if only_list_blocks:
      print(curblock_num,curblock,f.tell(),curblocksize)
      found = False
        
    
    if found:
      blocksize = curblocksize
      offset = f.tell()
    else:
      f.seek(curblocksize, os.SEEK_CUR)
      blocksize_check = (np.fromfile(f,dtype=np.uint32,count=1))[0]
      if swap: blocksize_check = blocksize_check.byteswap()
      if (curblocksize != blocksize_check):
        print("something wrong",curblocksize,blocksize_check)
        sys.exit()
      curblock_num += 1
  f.close()
      
  if ((not found) and (not only_list_blocks)):
    return -1,-1
    
  if (not only_list_blocks):
    return offset,blocksize
 








# ----- read the INFO block ----- 

def get_info(curfilename,block,swap):
  # Open the file
  f = open (curfilename,'rb')  
  f.seek(0, os.SEEK_END)
  filesize = f.tell()
  f.seek(0, os.SEEK_SET)

  # Look for the 'INFO' field
  found = False
  while ((not found) and (f.tell()<filesize)):
    f.seek(4, os.SEEK_CUR)
    label = f.read(4)
    if (label == 'INFO'):
      found = True
      break
    f.seek(8, os.SEEK_CUR)  
    curblocksize = (np.fromfile(f,dtype=np.uint32,count=1))[0]
    if swap:
      curblocksize = curblocksize.byteswap()
    f.seek(curblocksize+4, os.SEEK_CUR)

  # If the INFO block is present, read it
  if (found == True):
    dummy = f.read(8)
    blcksz = np.fromfile(f,dtype=np.int32,count=1)
    if swap:
      blcksz = blcksz.byteswap()
    nfields = blcksz//(12+4*7)
    myname = 'XXXX'
    for cnt in range(nfields):
      myname = f.read(4)
      mytype = f.read(8)
      myinfo = np.fromfile(f,dtype=np.uint32,count=7)
      if swap:
        myinfo = myinfo.byteswap()
      if (block == 'INFO'):
        print(myname,':\t',mytype,' ',myinfo)
      cnt += 1
      if (myname == block):
        f.close()
        return mytype,myinfo[0],myinfo[1:]
    if (block == 'INFO'):
      return 0,0,0
    print('The block you are asking for is not present in the INFO block')
    sys.exit()

  # If there is no INFO block, then assume the format
  if (block == 'MASS'):
    f.close()
    return 'FLOAT   ',1,np.array([1,1,1,1,1,1])
  elif (block in ['POS ','VEL ']):
    f.close()
    return 'FLOATN  ',3,np.array([1,1,1,1,1,1])
  elif (block == 'ID  '):
    f.close()
    return 'LONG    ',1,np.array([1,1,1,1,1,1])
  else:
    print('In this snapshot there is no INFO field and you are asking for a block which is not one of the following: MASS, POS, VEL, ID')
    sys.exit()
  




# ----- read data block -----
def read_block(filename, block):
  
  if os.path.exists(filename):
    curfilename = filename
  elif os.path.exists(filename+".0"):
    curfilename = filename+".0"
  else:
    print("file not found:", filename)
    sys.exit()
  
  head = snapshot_header(curfilename)
  swap = head.swap
  npart = head.npart # number of particles in this file
  nall = head.nall # total number of particles
  massarr = head.massarr
  filenum = head.filenum
  redshift = head.redshift
  time = head.time
 
  if (block == 'HEAD'):
    return head # the program stops here

  del head

  # read the INFO block and get the information on the fields. 
  # If there is no INFO block, then the usual formats are assumed (POS, VEL, MASS, ID)
  (datatype,dimdata,data_for_type) = get_info(curfilename,block,swap)
  if (datatype == 0):
    return 0 # The INFO block has been printed and now the program stops

  if (datatype == 'FLOAT   '):
    dt = np.dtype(np.float32)    
  elif (datatype == 'FLOATN  '):
    dt = np.dtype((np.float32,dimdata))    
  if (datatype == 'DOUBLE   '):
    dt = np.dtype(np.float64)    
  elif (datatype == 'DOUBLEN  '):
    dt = np.dtype((np.float64,dimdata))    
  elif (datatype == 'LONG    '):
    dt = np.dtype(np.uint32)    
  elif (datatype == 'LONGN   '):
    dt = np.dtype((np.uint32,dimdata))    
  elif (datatype == 'LLONG   '):
    dt = np.dtype(np.uint64)    
  elif (datatype == 'LLONGN  '):
    dt = np.dtype((np.uint64,dimdata))    

    
  allpartnum = np.int64(0)
  species_offset = np.zeros(6,np.int64)
  for j in range(6):
    species_offset[j] = allpartnum
    if data_for_type[j]:
      allpartnum += nall[j]

  for i in range(filenum): # main loop over files
    if filenum>1:
      curfilename = filename+"."+str(i)
      
    if i>0:
      head = snapshot_header(curfilename)
      npart = head.npart  
      del head
      
    curpartnum = np.int32(0)
    cur_species_offset = np.zeros(6,np.int64)
    for j in range(6):
      cur_species_offset[j] = curpartnum
      if data_for_type[j]:
        curpartnum += npart[j]
    
    # get the offset and the lenght of the block
    offset,blocksize = find_block(curfilename,swap,block,1)
    if (offset == -1 & i == 0):
      print("Error: block not found")
      sys.exit()
    elif ((offset == -1) & (i > 0)):
      return data # the program stops here


    if i==0: # fix data type for ID if long IDs are used
      if block=="ID  ":
        if blocksize == np.dtype(dt).itemsize*curpartnum * 2:
          dt = np.uint64 
        
    if np.dtype(dt).itemsize*curpartnum != blocksize:
      print("something wrong with blocksize! expected =",np.dtype(dt).itemsize*curpartnum,"actual =",blocksize)
      print("itemsize,curpartnum = ",np.dtype(dt).itemsize,curpartnum)
      sys.exit()
    
    f = open(curfilename,'rb')
    f.seek(offset, os.SEEK_CUR)  
    curdat = np.fromfile(f,dtype=dt,count=curpartnum) # read data
    f.close()  
    if swap:
      curdat.byteswap(True)  
      
    if i==0:
      data = np.ones(allpartnum,dt)
    
    for j in range(6):
      if data_for_type[j]:
        if block=="MASS" and massarr[j]>0: # add mass block for particles for which the mass is specified in the snapshot header
          data[species_offset[j]:species_offset[j]+npart[j]] = massarr[j]
        else:
          data[species_offset[j]:species_offset[j]+npart[j]] = curdat[cur_species_offset[j]:cur_species_offset[j]+npart[j]]
        species_offset[j] += npart[j]

    del curdat

  return data
  













# ----- list all data blocks in a format 2 snapshot file -----

def list_format2_blocks(filename):
  if os.path.exists(filename):
    curfilename = filename
  elif os.path.exists(filename+".0"):
    curfilename = filename+".0"
  else:
    print("file not found:", filename)
    sys.exit()
  
  head = snapshot_header(curfilename)
  format = head.format
  swap = head.swap
  del head
  
  print('GADGET FORMAT ',format)
  if (format != 2):
    print("#   OFFSET   SIZE")
  else:            
    print("#   BLOCK   OFFSET   SIZE")
  print("-------------------------")
  
  find_block(curfilename, format, swap, "XXXX", 0, only_list_blocks=True)
  
  print("-------------------------")

def read_gadget_header(filename):
  if os.path.exists(filename):
    curfilename = filename
  elif os.path.exists(filename+".0"):
    curfilename = filename+".0"
  else:
    print("file not found:", filename)
    sys.exit()

  head=snapshot_header(curfilename)
  print('npar=',head.npart)
  print('nall=',head.nall)
  print('a=',head.time)
  print('z=',head.redshift)
  print('masses=',head.massarr*1e10,'Msun/h')
  print('boxsize=',head.boxsize,'kpc/h')
  print('filenum=',head.filenum)
  print('cooling=',head.cooling)
  print('Omega_m,Omega_l=',head.omega_m,head.omega_l)
  print('h=',head.hubble,'\n')
  
  rhocrit=2.77536627e11 #h**2 M_sun/Mpc**3
  rhocrit=rhocrit/1e9 #h**2M_sun/kpc**3
  
  Omega_DM=head.nall[1]*head.massarr[1]*1e10/(head.boxsize**3*rhocrit)
  print('DM mass=',head.massarr[1]*1e10,'Omega_DM=',Omega_DM)
  if head.nall[2]>0 and head.massarr[2]>0:
    Omega_NU=head.nall[2]*head.massarr[2]*1e10/(head.boxsize**3*rhocrit)
    print('NU mass=',head.massarr[2]*1e10,'Omega_NU=',Omega_NU)
    print('Sum of neutrino masses=',Omega_NU*head.hubble**2*94.1745,'eV')



# This script is an example on how to run the void finder on a Gadget snapshot
import numpy as np
import readgadget
import MAS_library as MASL
import void_library as VL

################################# INPUT ######################################
snapshot = '/simons/scratch/fvillaescusa/pdf_information/fiducial/0/snapdir_004/snap_004'

# density field parameters
grid = 768
MAS  = 'PCS'

# void finder parameters
Radii      = np.array([5, 7, 9, 11, 13, 15, 17, 19, 21, 23, 25, 27,
                       29, 31, 33, 35, 37, 39, 41], dtype=np.float32) #Mpc/h
threshold  = -0.7
threads1   = 16
threads2   = 4
void_field = True
##############################################################################

# read snapshot head and obtain BoxSize, Omega_m and Omega_L
head    = readgadget.header(snapshot)
BoxSize = head.boxsize/1e3  #Mpc/h

# read particle positions
pos = readgadget.read_block(snapshot,"POS ",ptype=1)/1e3 #Mpc/h

# compute density field
delta = np.zeros((grid, grid, grid), dtype=np.float32)
MASL.MA(pos, delta, BoxSize, MAS)
delta /= np.mean(delta, dtype=np.float64);  delta -= 1.0

# find the voids
V = VL.void_finder(delta, BoxSize, threshold, Radii, 
                   threads1, threads2, void_field=void_field)

# void properties
void_pos    = V.void_pos     #Mpc/h
void_radius = V.void_radius  #Mpc/h
VSF_R       = V.Rbins        #VSF bins in radius
VSF         = V.void_vsf     #VSF
if void_field:
    void_field = V.void_field  #array with 0. Cells belonging to voids have 1





# The purpose of this library is to take an unsorted set of particles
# positions and return a structure that can be be used to quickly find
# particles in a given region of the space
from __future__ import print_function
import numpy as np 
import time,sys,os
cimport numpy as np
cimport cython
from libc.math cimport sqrt,pow,sin,log10,abs,exp,log,log10
from cpython cimport bool


# This routine reads an array with particle positions and computes the associated
# index of it:  index = dims2*i + dims*j + k
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cpdef indexes_3D_cube(np.float32_t[:,:] pos, float BoxSize, float cell_size):

    cdef long particles, dims2, i
    cdef int dims, ii, jj, kk
    cdef np.int64_t[:] index
    cdef float factor

    # find the total number of halos in the catalogue
    particles = pos.shape[0]
    dims = <int>(BoxSize/cell_size) 
    dims2 = dims*dims
    factor = dims*1.0/BoxSize

    # each particle is characterized by its index = dims2*ii + dims*jj + kk
    # index_sorted is just the index array sorted
    index = np.zeros(particles, dtype=np.int64)

    # for each halo find its index
    for i in range(particles):
        ii = <int>(pos[i,0]*factor)
        if ii==dims:  ii = 0
        jj = <int>(pos[i,1]*factor)
        if jj==dims:  jj = 0
        kk = <int>(pos[i,2]*factor)
        if kk==dims:  kk = 0

        index[i] = ii*dims2 + jj*dims + kk

    return np.asarray(index)

# This routine takes as input an array with the positions of particles
# and returns an array with the particle positions sorted together with
# an array that specifies how many particles are in each cell. We assume
# periodic boundary conditions
# pos ----------> positions of the particles
# Boxsize ------> size of the box
# cell_size ----> size of the cells to organize the data
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
class sort_3D_pos:
    def __init__(self, np.float32_t[:,:] pos, float BoxSize, float cell_size,
            bool return_offset=True, bool return_indexes=True):

        print('Sorting particles... ',end='');  start = time.time()
        cdef long particles, dims2, dims3, i, number
        cdef int dims, ii, jj, kk
        cdef np.int64_t[:] index, index_sorted, offset, indexes
        cdef np.float32_t[:,:] pos_sorted
        cdef float factor

        # find the total number of halos in the catalogue
        particles = pos.shape[0]
        dims = <int>(BoxSize/cell_size) 
        dims2, dims3 = dims*dims, dims*dims*dims
        factor = dims*1.0/BoxSize

        # each particle is characterized by its index = dims2*ii + dims*jj + kk
        # index_sorted is just the index array sorted
        index        = np.zeros(particles,   dtype=np.int64)
        index_sorted = np.zeros(particles,   dtype=np.int64)

        # for each halo find its index
        for i in range(particles):
            ii = <int>(pos[i,0]*factor)
            if ii==dims:  ii = 0
            jj = <int>(pos[i,1]*factor)
            if jj==dims:  jj = 0
            kk = <int>(pos[i,2]*factor)
            if kk==dims:  kk = 0

            index[i] = ii*dims2 + jj*dims + kk

        # find the positions where the array index will be sorted
        # e.g. index = [1 0 4 9 3 2] ---> indexes = [1 0 5 4 2 3]
        # so index[indexes] = [0 1 2 3 4 9] 
        indexes = np.argsort(index)

        if return_indexes:  self.indexes = np.asarray(indexes)

        # get the new arrays with the sorted positions and the 
        # sorted indexes
        pos_sorted   = np.empty((particles,3), dtype=np.float32)
        index_sorted = np.empty(particles,     dtype=np.int64)
        for i in range(particles):
            number = indexes[i]
            pos_sorted[i,0] = pos[number,0]
            pos_sorted[i,1] = pos[number,1]
            pos_sorted[i,2] = pos[number,2]
            index_sorted[i] = index[number]
        del index
        self.pos_sorted = np.asarray(pos_sorted)

        # obtain the offset array that stores where the particles
        # in each cell begin and end
        if return_offset:
            offset = np.zeros(dims3+1, dtype=np.int64) - 1 #put -1 to empty cells 
            number = index_sorted[0]
            offset[index_sorted[0]] = 0
            for i in range(particles):
                if index_sorted[i]!=number:
                    offset[index_sorted[i]] = i
                    number = index_sorted[i]
            offset[dims3] = particles

            number = offset[dims3]
            for i in range(dims3,-1, -1):
                if offset[i]==-1:  offset[i] = number
                else:              number    = offset[i]

            self.offset = np.asarray(offset)

        print('Done : %.3f seconds'%(time.time()-start))



# This routine takes two arrays with positions as inputs. It computes the
# distances between the first array to the second one keeping only the
# closest distances
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
cpdef nearby_particles(np.float32_t[:,:] pos1, np.float32_t[:,:] pos2, 
    np.float32_t[:] radii2, np.int64_t[:] indexes_part, float BoxSize, float cell_size):

    print('Finding nearby halos...');  start = time.time()
    cdef np.float32_t[:,:] pos1_sorted, pos2_sorted
    cdef np.float32_t[:] radii2_sorted
    cdef np.int64_t[:] offset1, offset2, indexes1, indexes2
    cdef long i, j, particles1, particles2, index, dims2, particles_counted
    cdef float x, y, z, dist, factor, dx, dy, dz
    cdef int dims, ii, jj, kk, iii, jjj, kkk, i_index, j_index, k_index
    cdef list halo_numbers

    # find the number of particles in each array
    particles1 = pos1.shape[0]
    particles2 = pos2.shape[0]

    # find the total number of halos in the catalogue
    dims   = <int>(BoxSize/cell_size) 
    dims2  = dims*dims
    factor = dims*1.0/BoxSize

    # sort the positions of pos1 and pos2
    pos1_sorted, offset1, indexes1 = sort_3D_pos(pos1, BoxSize, cell_size)
    pos2_sorted, offset2, indexes2 = sort_3D_pos(pos2, BoxSize, cell_size)

    radii2_sorted = np.empty(radii2.shape[0], dtype=np.float32)
    for i in range(radii2.shape[0]):
        radii2_sorted[i] = radii2[indexes2[i]]

    halo_numbers = []
    for i in range(particles1):
        halo_numbers.append([])

    # do a loop over the particles2
    particles_counted = 0
    for i in range(particles1):
        x, y, z = pos1_sorted[i,0], pos1_sorted[i,1], pos1_sorted[i,2]

        ii = <int>(pos1_sorted[i,0]*factor)
        jj = <int>(pos1_sorted[i,1]*factor)
        kk = <int>(pos1_sorted[i,2]*factor)
        
        for iii in range(ii-1, ii+2):
            i_index = (iii+dims)%dims
            for jjj in range(jj-1, jj+2):
                j_index = (jjj+dims)%dims
                for kkk in range(kk-1, kk+2):
                    k_index = (kkk+dims)%dims

                    index = dims2*i_index + dims*j_index + k_index
                    particles_counted += (offset2[index+1] - offset2[index])

                    for j in range(offset2[index], offset2[index+1]):
                        dx = pos2_sorted[j,0]-x
                        if dx<-BoxSize/2.0:  dx += BoxSize
                        if dx>BoxSize/2.0:   dx -= BoxSize

                        dy = pos2_sorted[j,1]-y
                        if dy<-BoxSize/2.0:  dy += BoxSize
                        if dy>BoxSize/2.0:   dy -= BoxSize

                        dz = pos2_sorted[j,2]-z
                        if dz<-BoxSize/2.0:  dz += BoxSize
                        if dz>BoxSize/2.0:   dz -= BoxSize

                        dist = dx*dx + dy*dy + dz*dz

                        if dist<(radii2_sorted[j]*radii2_sorted[j]*1.3) or dist<0.100:
                            halo_numbers[i].append(indexes_part[indexes2[j]])

    print('Time taken = %.3f seconds'%(time.time()-start))
    print('Particles counted in = %ld'%particles_counted)

    return halo_numbers, indexes1










from platform import machine, system

import numpy
from Cython.Build import cythonize
from setuptools import Extension, find_packages, setup


is_mac = system() == "Darwin"
is_arm = machine() == "arm64"
is_m1 = is_mac and is_arm

arch_flag = "-mcpu=apple-m1" if is_m1 else "-march=native"
omp_flag = "-Xpreprocessor -fopenmp" if is_m1 else "-fopenmp"
extra_compile_args = ["-O3", "-ffast-math"]
extra_compile_args_omp = extra_compile_args.copy()
extra_compile_args_omp.append(omp_flag)
extra_link_args = [omp_flag]

ext_modules = [
    Extension(
        "MAS_library.MAS_library",
        ["library/MAS_library/MAS_library.pyx", "library/MAS_library/MAS_c.c"],
        extra_compile_args=extra_compile_args_omp,
        extra_link_args=extra_link_args,
        libraries=["m"],
    ),
    Extension(
        "Pk_library.Pk_library",
        ["library/Pk_library/Pk_library.pyx"],
        extra_compile_args=extra_compile_args_omp,
    ),
    Extension(
        "Pk_library.bispectrum_library", ["library/Pk_library/bispectrum_library.pyx"]
    ),
    Extension(
        "MAS_library.field_properties", ["library/MAS_library/field_properties.pyx"]
    ),
    Extension(
        "redshift_space_library.redshift_space_library",
        ["library/redshift_space_library/redshift_space_library.pyx"],
    ),
    Extension(
        "smoothing_library.smoothing_library",
        ["library/smoothing_library/smoothing_library.pyx"],
        extra_compile_args=extra_compile_args_omp,
        extra_link_args=extra_link_args,
        libraries=["m"],
    ),
    Extension(
        "void_library.void_library",
        [
            "library/void_library/void_library.pyx",
            "library/void_library/void_openmp_library.c",
        ],
        extra_compile_args=extra_compile_args_omp,
        extra_link_args=extra_link_args,
        libraries=["m"],
    ),
    Extension(
        "integration_library.integration_library",
        [
            "library/integration_library/integration_library.pyx",
            "library/integration_library/integration.c",
            "library/integration_library/runge_kutta.c",
        ],
        extra_compile_args=extra_compile_args,
    ),
    Extension(
        "density_field_library.density_field_library",
        ["library/density_field_library/density_field_library.pyx"],
    ),
    Extension(
        "sorting_library.sorting_library",
        ["library/sorting_library/sorting_library.pyx"],
        extra_compile_args=extra_compile_args,
    ),
    Extension("HI_library.HI_library", ["library/HI_library/HI_library.pyx"]),
    Extension(
        "HI_clusters_library.HI_clusters_library",
        ["library/HI_clusters_library/HI_clusters_library.pyx"],
    ),
]


with open("README.md", "r") as f:
    documentation = f.read()

setup(
    name="Pylians",
    version="0.12",
    author="Francisco Villaescusa-Navarro",
    author_email="villaescusa.francisco@gmail.com",
    description="Python libraries for the analysis of numerical simulations",
    url="https://github.com/franciscovillaescusa/Pylians3",
    license="MIT",
    long_description_content_type="text/markdown",
    long_description=documentation,
    packages=find_packages(where="library/"),
    ext_modules=cythonize(
        ext_modules,
        compiler_directives={"language_level": "3"},
        include_path=[
            "library/MAS_library/",
            "library/void_library/",
            "library/integration_library/",
        ],
    ),
    include_dirs=[numpy.get_include()],
    install_requires=[
        "h5py",
        "pyfftw; platform_system!='Darwin' and platform_machine!='arm64'",
        "scipy",
        "hdf5plugin",
        "Cython<3.0.0",
    ],
    package_dir={"": "library/"},
    py_modules=[
        "bias_library",
        "CAMB_library",
        "correlation_function_library",
        "cosmology_library",
        "HI_image_library/HI_image_library",
        "HOD_library",
        "IM_library",
        "mass_function_library",
        "plotting_library",
        "readfof",
        "readgadget",
        "readsnapHDF5",
        "readsnap",
        "readsnap2",
        "readsnap_mpi",
        "readsubf",
        "routines",
        "units_library",
    ],
)



import numpy as np
#import scipy.integrate as si
import integration_library as IL
import readfof, readsubf
import os, sys
import units_library as UL


############################ CONSTANTS ############################
pi       = np.pi
rho_crit = (UL.units()).rho_crit #h^2 Msun/Mpc^3
deltac   = 1.686
###################################################################


# This is the main driver routine to compute the halo mass function from several
# fitting formulae.
# k_in ---------> input k
# Pk_in --------> input Pk
# OmegaM -------> value of Omega_m
# Masses -------> array with the masses where to compute the HMF
# author -------> 'ST','Tinker','Tinker10','Crocce','Jenkins','Warren',
#                 'Watson','Watson_FoF','Angulo'
# bins ---------> In order to carry out the integral, we need a k-array sorted and
#                 equally spaced in log10. This sets the number of bins in that array
# z ------------> for 'Tinker', 'Tinker10' and Crocce
# delta --------> for 'Tinker' and 'Tinker10'
def MF_theory(k_in, Pk_in, OmegaM, Masses, author, bins=10000, z=0, delta=200.0):

    rhoM = rho_crit*OmegaM

    # sort the input k-array
    indexes = np.argsort(k_in)
    k_in = k_in[indexes];  Pk_in = Pk_in[indexes]
    
    # interpolate the input k-array into bins points
    k  = np.logspace(np.log10(k_in[0]), np.log10(k_in[-1]), bins)
    Pk = np.interp(np.log(k), np.log(k_in), Pk_in)
    
    if author=='ST':            MF = ST_mass_function(k, Pk, rhoM, Masses)
    elif author=="Tinker":      MF = Tinker_mass_function(k, Pk, rhoM, Masses,
                                                          z, delta)
    elif author=="Tinker10":    MF = Tinker_2010_mass_function(k, Pk, rhoM, Masses,
                                                               z, delta)
    elif author=="Crocce":      MF = Crocce_mass_function(k, Pk, rhoM, Masses, z)
    elif author=="Jenkins":     MF = Jenkins_mass_function(k, Pk, rhoM, Masses)
    elif author=="Warren":      MF = Warren_mass_function(k, Pk, rhoM, Masses)
    elif author=="Watson":      MF = Watson_mass_function(k, Pk, rhoM, Masses)
    elif author=="Watson_FOF":  MF = Watson_mass_function_FoF(k, Pk, rhoM, Masses)
    elif author=="Angulo":      MF = Angulo_subhalos_mass_function(k, Pk, rhoM, Masses)
    else:  raise Exception("%s model not implemented!!!"%author)

    return MF

    

"""
###### old method to compute integrals ######
#derivative function for integrating sigma(R)
def deriv_sigma(y,x,k,Pk,R):
    Pkp=np.interp(np.log10(x),np.log10(k),np.log10(Pk)); Pkp=10**Pkp
    kR=x*R
    W=3.0*(np.sin(kR)-kR*np.cos(kR))/kR**3 
    return [x**2*Pkp*W**2]

#this function computes sigma(R)
def sigma_old(k,Pk,R):
    k_limits=[k[0],k[-1]]; yinit=[0.0]

    I=si.odeint(deriv_sigma,yinit,k_limits,args=(k,Pk,R),
                rtol=1e-8,atol=1e-8,
                mxstep=10000000)[1][0]/(2.0*pi**2)

    return np.sqrt(I)
"""

def sigma(k,Pk,R):
    yinit = np.array([0.0], dtype=np.float64)
    eps   = 1e-13  #change this for higher/lower accuracy
    h1    = 1e-12
    hmin  = 0.0
    
    W   = 3.0*(np.sin(k*R) - k*R*np.cos(k*R))/(k*R)**3
    Pk1 = Pk*W**2*k**2/(2.0*pi**2)
    
    return np.sqrt(IL.odeint(yinit, k[0], k[-1], eps,
                             h1, hmin, np.log10(k), Pk1,
                             'sigma', verbose=False)[0])

#this function computes the derivate of sigma(M) wrt M
def dSdM(k, Pk, rhoM, M):

    R1=(3.0*M/(4.0*pi*rhoM))**(1.0/3.0)
    s1=sigma(k,Pk,R1)

    M2=M*1.0001
    R2=(3.0*M2/(4.0*pi*rhoM))**(1.0/3.0)
    s2=sigma(k,Pk,R2)

    return (s2-s1)/(M2-M)

##############################################################################
#This function computes the Sheth-Tormen mass function. Returns dn/dM
#If the mass function wants to be computed in a given mass bins, use Masses
def ST_mass_function(k, Pk, rhoM, Masses):

    dndM = np.zeros(Masses.shape[0], dtype=np.float64)
        
    for i,M in enumerate(Masses):
        R   = (3.0*M/(4.0*pi*rhoM))**(1.0/3.0)
        nu  = (1.686/sigma(k,Pk,R))**2
        nup = 0.707*nu

        dndM[i]=-2.0*(rhoM/M)*dSdM(k,Pk,rhoM,M)/sigma(k,Pk,R)
        dndM[i]*=0.3222*(1.0+1.0/nup**0.3)*np.sqrt(0.5*nup)
        dndM[i]*=np.exp(-0.5*nup)/np.sqrt(pi)

    return dndM

##############################################################################
#This function computes the Tinker mass function. Returns dn/dM
#If the mass function wants to be computed in a given mass bins, use Masses
def Tinker_mass_function(k, Pk, rhoM, Masses, z, delta=200.0):

    alpha=10**(-(0.75/np.log10(delta/75.0))**1.2)

    D=[200.0, 300.0, 400.0, 600.0, 800.0, 1200.0, 1600.0, 2400.0, 3600.0]
    A=[0.186, 0.200, 0.212, 0.218, 0.248, 0.255,  0.260,  0.260,  0.260]
    a=[1.47,  1.52,  1.56,  1.61,  1.87,  2.13,   2.30,   2.53,   2.66]
    b=[2.57,  2.25,  2.05,  1.87,  1.59,  1.51,   1.46,   1.44,   1.41]
    c=[1.19,  1.27,  1.34,  1.45,  1.58,  1.80,   1.97,   2.24,   2.44]
    D=np.array(D); A=np.array(A); a=np.array(a); b=np.array(b); c=np.array(c)

    A=np.interp(delta,D,A)
    a=np.interp(delta,D,a)
    b=np.interp(delta,D,b)
    c=np.interp(delta,D,c)

    # this is for R200_critical with OmegaM=0.2708
    A*=(1.0+z)**(-0.14)
    a*=(1.0+z)**(-0.06)
    b*=(1.0+z)**(-alpha)

    print('Delta=',delta)
    print('A=',A)
    print('a=',a)
    print('b=',b)
    print('c=',c)
    
    dndM = np.zeros(Masses.shape[0], dtype=np.float64)
        
    for i,M in enumerate(Masses):
        R   =(3.0*M/(4.0*pi*rhoM))**(1.0/3.0)
        s   = sigma(k,Pk,R)
        f_s = A*((b/s)**(a)+1.0)*np.exp(-c/s**2)

        dndM[i] = -(rhoM/M)*dSdM(k,Pk,rhoM,M)*f_s/s

    return dndM

##############################################################################
def Tinker_2010_mass_function(k, Pk, rhoM, Masses, z, delta=200.0):

    if delta!=200.0:
        raise Exception("only implemented delta=200, please update library")

    alpha  = 0.368
    beta0  = 0.589;       beta  = beta0*(1.0+z)**0.20
    gamma0 = 0.864;       gamma = gamma0*(1.0+z)**(-0.01)
    phi0   = -0.729;      phi   = phi0*(1.0+z)**(-0.08)
    eta0   = -0.243;      eta   = eta0*(1.0+z)**0.27
    
    dndM = np.empty(Masses.shape[0], dtype=np.float64)
        
    for i,M in enumerate(Masses):
        R = (3.0*M/(4.0*pi*rhoM))**(1.0/3.0)
        nu = 1.686/sigma(k,Pk,R)
        fnu = alpha*(1.0+(beta*nu)**(-2.0*phi))*nu**(2.0*eta)*\
            np.exp(-gamma*nu**2/2.0)

        #Mp  = M*1.01;  Rp = (3.0*Mp/(4.0*pi*rhoM))**(1.0/3.0)
        #nup = 1.686/sigma(k,Pk,Rp)
        #dlnnu_dlnM = np.log(nup-nu)/np.log(Mp-M)

        #dndM[i] = -(rhoM/M**2)*(nu*fnu)*dlnnu_dlnM
        dndM[i] = -(rhoM/M)*dSdM(k,Pk,rhoM,M)/sigma(k,Pk,R)*(nu*fnu)

    return dndM
    


##############################################################################
#This function computes the Crocce mass function. Returns dn/dM
#If the mass function wants to be computed in a given mass bins, use Masses
def Crocce_mass_function(k, Pk, rhoM, Masses, z):

    A=0.58*(1.0+z)**(-0.13) 
    a=1.37*(1.0+z)**(-0.15)
    b=0.3*(1.0+z)**(-0.084)
    c=1.036*(1.0+z)**(-0.024)
    
    dndM = np.zeros(Masses.shape[0], dtype=np.float64)
        
    for i,M in enumerate(Masses):
        R   = (3.0*M/(4.0*pi*rhoM))**(1.0/3.0)
        s   = sigma(k,Pk,R)
        f_s = A*(s**(-a)+b)*np.exp(-c/s**2)

        dndM[i]=-(rhoM/M)*dSdM(k,Pk,rhoM,M)*f_s/s

    return dndM

##############################################################################
#This function computes the Jenkins mass function. Returns dn/dM
#If the mass function wants to be computed in a given mass bins, use Masses
def Jenkins_mass_function(k, Pk, rhoM, Masses):

    A=0.315
    b=0.61
    c=3.8
    
    dndM = np.zeros(Masses.shape[0], dtype=np.float64)
        
    for i,M in enumerate(Masses):
        R   = (3.0*M/(4.0*pi*rhoM))**(1.0/3.0)
        s   = sigma(k,Pk,R)
        f_s = A*np.exp(-np.absolute(np.log(1.0/s)+b)**c)

        dndM[i] = -(rhoM/M)*dSdM(k,Pk,rhoM,M)*f_s/s

    return dndM

##############################################################################
#This function computes the Warren mass function. Returns dn/dM
#If the mass function wants to be computed in a given mass bins, use Masses
def Warren_mass_function(k, Pk, rhoM, Masses):

    A=0.7234
    a=1.625
    b=0.2538
    c=1.1982
    
    dndM = np.zeros(Masses.shape[0], dtype=np.float64)
        
    for i,M in enumerate(Masses):
        R   = (3.0*M/(4.0*pi*rhoM))**(1.0/3.0)
        s   = sigma(k,Pk,R)
        f_s = A*(s**(-a)+b)*np.exp(-c/s**2)

        dndM[i] = -(rhoM/M)*dSdM(k,Pk,rhoM,M)*f_s/s

    return dndM

##############################################################################
#This function computes the Warren mass function for FoF halos. Returns dn/dM
#If the mass function wants to be computed in a given mass bins, use Masses
def Watson_mass_function_FoF(k, Pk, rhoM, Masses):

    A=0.282
    a=2.163
    b=1.406
    c=1.210
    
    dndM = np.zeros(Masses.shape[0], dtype=np.float64)
        
    for i,M in enumerate(Masses):
        R   = (3.0*M/(4.0*pi*rhoM))**(1.0/3.0)
        s   = sigma(k,Pk,R)
        f_s = A*((b/s)**a+1.0)*np.exp(-c/s**2)

        dndM[i] = -(rhoM/M)*dSdM(k,Pk,rhoM,M)*f_s/s

    return dndM

##############################################################################
#This function computes the Warren mass function. Returns dn/dM
#If the mass function wants to be computed in a given mass bins, use Masses
def Watson_mass_function(k, Pk, rhoM, Masses):

    delta=200.0
    OmegaM =rhoM/rho_crit

    A=0.194
    a=1.805
    b=2.267
    c=1.287

    dndM = np.zeros(Masses.shape[0], dtype=np.float64)
        
    for i,M in enumerate(Masses):
        R   = (3.0*M/(4.0*pi*rhoM))**(1.0/3.0)
        s   = sigma(k,Pk,R)
        f_s = A*(s**(-a)+b)*np.exp(-c/s**2)

        factor=np.exp(0.023*(delta/178.0-1.0))
        factor*=(delta/178.0)**(-0.456*OmegaM-0.139)
        factor*=np.exp(0.072*(1-delta/178.0)/s**2.130)

        f_s*=factor

        dndM[i] = -(rhoM/M)*dSdM(k,Pk,rhoM,M)*f_s/s

    return dndM

#############################################################################
#This function computes the Warren mass function. Returns dn/dM
#If the mass function wants to be computed in a given mass bins, use Masses
def Angulo_subhalos_mass_function(k, Pk, rhoM, Masses):
    
    dndM = np.zeros(Masses.shape[0], dtype=np.float64)
        
    for i,M in enumerate(Masses):
        R   = (3.0*M/(4.0*pi*rhoM))**(1.0/3.0)
        s   = sigma(k,Pk,R)
        f_s = 0.265*(1.675/s+1.0)**1.9*np.exp(-1.4/s**2)

        dndM[i] = -(rhoM/M)*dSdM(k,Pk,rhoM,M)*f_s/s

    return dndM


#############################################################################
#This functions computes the halo mass function (dn/dM (M)) of a given object
#and write the results to a file. The arguments are:
#groups_fname ---> folder containing the halos/subhalos files
#groups_number --> number of the snapshot (e.g. 1, 2, 22...)
#f_out ----------> the name of the output file
#min_mass -------> the lower limit of the mass function (units of 1e10 Msun/h)
#min_mass -------> the upper limit of the mass function (units of 1e10 Msun/h)
#bins -----------> the number of bins used to compute the mass function
#BoxSize --------> Size of the simulation box (in Mpc/h)
#obj ------------> the object over which compute the mass function:
#   *FoF halos -----------> 'FoF'
#   *SO 200xmean halos ---> 'halos_m200'
#   *subhalos -------- ---> 'subhalos'
#long_IDs_flag -> flag for long IDs (True or False)
#SFR_flag ------> flag for SFR (True or False)
#When using FoF halos their masses are corrected to account for sampling effects
#if min_mass=None the code will take the minimum mass of the halo catalogue
#the same applies to max_mass
def mass_function(groups_fname,groups_number,obj,BoxSize,bins,f_out,
                  min_mass=None,max_mass=None,
                  long_ids_flag=True,SFR_flag=False):

    #bins_mass=np.logspace(np.log10(min_mass),np.log10(max_mass),bins+1)
    #mass_mean=10**(0.5*(np.log10(bins_mass[1:])+np.log10(bins_mass[:-1])))
    #dM=bins_mass[1:]-bins_mass[:-1]

    if obj=='FoF':
        #read FoF halos information
        fof=readfof.FoF_catalog(groups_fname,groups_number,
                                long_ids=long_ids_flag,swap=False,SFR=SFR_flag)
        F_pos=fof.GroupPos/1e3        #positions in Mpc/h
        F_mass=fof.GroupMass*1e10     #masses in Msun/h
        F_part=fof.GroupLen           #number particles belonging to the group
        F_Mpart=F_mass[0]/F_part[0]   #mass of a single particle in Msun/h
        del fof

        #some verbose
        print('\nNumber of FoF halos=',len(F_pos))
        print('%f < X_fof < %f'%(np.min(F_pos[:,0]),np.max(F_pos[:,0])))
        print('%f < Y_fof < %f'%(np.min(F_pos[:,1]),np.max(F_pos[:,1])))
        print('%f < Z_fof < %f'%(np.min(F_pos[:,2]),np.max(F_pos[:,2])))
        print('%e < M_fof < %e\n'%(np.min(F_mass),np.max(F_mass)))

        #Correct the masses of the FoF halos
        F_mass=F_Mpart*(F_part*(1.0-F_part**(-0.6)))

        #compute the minimum and maximum mass
        if min_mass==None:
            min_mass=np.min(F_mass)
        if max_mass==None:
            max_mass=np.max(F_mass)

        print('M_min = %e'%(min_mass))
        print('M_max = %e\n'%(max_mass))

        #find the masses and mass intervals
        bins_mass=np.logspace(np.log10(min_mass),np.log10(max_mass),bins+1)
        mass_mean=10**(0.5*(np.log10(bins_mass[1:])+np.log10(bins_mass[:-1])))
        dM=bins_mass[1:]-bins_mass[:-1]

        #compute the number of halos within each mass interval
        number=np.histogram(F_mass,bins=bins_mass)[0]
        print(number); print(np.sum(number,dtype=np.float64))

    elif obj=='halos_m200':
        #read CDM halos information
        halos=readsubf.subfind_catalog(groups_fname,groups_number,
                                       group_veldisp=True,masstab=True,
                                       long_ids=True,swap=False)

        H_pos=halos.group_pos/1e3               #positions in Mpc/h
        H_mass=halos.group_m_mean200*1e10       #masses in Msun/h
        H_radius_m=halos.group_r_mean200/1e3    #radius in Mpc/h
        del halos

        #some verbose
        print('\nNumber of halos=',len(H_pos))
        print('%f < X < %f'%(np.min(H_pos[:,0]),np.max(H_pos[:,0])))
        print('%f < Y < %f'%(np.min(H_pos[:,1]),np.max(H_pos[:,1])))
        print('%f < Z < %f'%(np.min(H_pos[:,2]),np.max(H_pos[:,2])))
        print('%e < M < %e\n'%(np.min(H_mass),np.max(H_mass)))

        #compute the minimum and maximum mass
        if min_mass==None:
            min_mass=np.min(H_mass)
        if max_mass==None:
            max_mass=np.max(H_mass)

        #compute the number of halos within each mass interval
        bins_mass=np.logspace(np.log10(min_mass),np.log10(max_mass),bins+1)
        mass_mean=10**(0.5*(np.log10(bins_mass[1:])+np.log10(bins_mass[:-1])))
        dM=bins_mass[1:]-bins_mass[:-1]

        number=np.histogram(H_mass,bins=bins_mass)[0]
        print(number); print(np.sum(number,dtype=np.float64))

    elif obj=='halos_c200':
        #read CDM halos information
        halos=readsubf.subfind_catalog(groups_fname,groups_number,
                                       group_veldisp=True,masstab=True,
                                       long_ids=True,swap=False)

        H_pos=halos.group_pos/1e3               #positions in Mpc/h
        H_mass=halos.group_m_crit200*1e10       #masses in Msun/h
        H_radius_m=halos.group_r_crit200/1e3    #radius in Mpc/h
        del halos

        #some verbose
        print('\nNumber of halos=',len(H_pos))
        print('%f < X < %f'%(np.min(H_pos[:,0]),np.max(H_pos[:,0])))
        print('%f < Y < %f'%(np.min(H_pos[:,1]),np.max(H_pos[:,1])))
        print('%f < Z < %f'%(np.min(H_pos[:,2]),np.max(H_pos[:,2])))
        print('%e < M < %e\n'%(np.min(H_mass),np.max(H_mass)))

        #compute the minimum and maximum mass
        if min_mass==None:
            min_mass=np.min(H_mass)
        if max_mass==None:
            max_mass=np.max(H_mass)

        #compute the number of halos within each mass interval
        bins_mass=np.logspace(np.log10(min_mass),np.log10(max_mass),bins+1)
        mass_mean=10**(0.5*(np.log10(bins_mass[1:])+np.log10(bins_mass[:-1])))
        dM=bins_mass[1:]-bins_mass[:-1]

        number=np.histogram(H_mass,bins=bins_mass)[0]
        print(number); print(np.sum(number,dtype=np.float64))

    elif obj=='subhalos':
        #read CDM halos information
        halos=readsubf.subfind_catalog(groups_fname,groups_number,
                                       group_veldisp=True,masstab=True,
                                       long_ids=True,swap=False)

        S_pos=halos.sub_pos/1e3            #positions in Mpc/h
        S_mass=halos.sub_mass*1e10         #masses in Msun/h
        del halos

        #some verbose
        print('\nNumber of subhalos=',len(S_pos))
        print('%f < X < %f'%(np.min(S_pos[:,0]),np.max(S_pos[:,0])))
        print('%f < Y < %f'%(np.min(S_pos[:,1]),np.max(S_pos[:,1])))
        print('%f < Z < %f'%(np.min(S_pos[:,2]),np.max(S_pos[:,2])))
        print('%e < M < %e\n'%(np.min(S_mass),np.max(S_mass)))

        #compute the minimum and maximum mass
        if min_mass==None:
            min_mass=np.min(S_mass)
        if max_mass==None:
            max_mass=np.max(S_mass)

        #compute the number of halos within each mass interval
        bins_mass=np.logspace(np.log10(min_mass),np.log10(max_mass),bins+1)
        mass_mean=10**(0.5*(np.log10(bins_mass[1:])+np.log10(bins_mass[:-1])))
        dM=bins_mass[1:]-bins_mass[:-1]

        number=np.histogram(S_mass,bins=bins_mass)[0]
        print(number); print(np.sum(number,dtype=np.float64))

    else:
        print('bad object type selected')
        sys.exit()

    MF=number/(dM*BoxSize**3)
    delta_MF=np.sqrt(number)/(dM*BoxSize**3)

    f=open(f_out,'w')
    for i in range(bins):
        f.write(str(mass_mean[i])+' '+str(MF[i])+' '+str(delta_MF[i])+'\n')
    f.close()

#############################################################################

#This functions computes the function f(sigma) see eq. 1 of Crocce et al 2010
#for a given object, and write the results to a file. The arguments are:
#groups_fname ---> folder containing the halos/subhalos files
#groups_number --> number of the snapshot (e.g. 1, 2, 22...)
#f_out ----------> the name of the output file
#min_mass -------> the lower limit of the mass function (units of 1e10 Msun/h)
#min_mass -------> the upper limit of the mass function (units of 1e10 Msun/h)
#bins -----------> the number of bins used to compute the mass function
#BoxSize --------> Size of the simulation box (in Mpc/h)
#obj ------------> the object over which compute the mass function:
#   *FoF halos -----------> 'FoF'
#   *SO 200xmean halos ---> 'halos_m200'
#Omega_M -------> This is the value of Omega_M = Omega_CDM+Omega_B(+Omega_nu)
#k, Pk ---------> Those are the values of the DM P(k) or the CDM P(k) from CAMB
#When using FoF halos their masses are corrected to account for sampling effects

def mass_function_fsigma(groups_fname,groups_number,f_out,min_mass,max_mass,
                         bins,BoxSize,obj,Omega_M,k,Pk):

    rhoM=rho_crit*Omega_M
    bins_mass=np.logspace(np.log10(min_mass),np.log10(max_mass),bins+1)
    mass_mean=10**(0.5*(np.log10(bins_mass[1:])+np.log10(bins_mass[:-1])))

    if obj=='FoF':
        #read FoF halos information
        fof=readfof.FoF_catalog(groups_fname,groups_number,
                                long_ids=True,swap=False)
        F_pos=fof.GroupPos/1e3        #positions in Mpc/h
        F_mass=fof.GroupMass*1e10     #masses in Msun/h
        F_part=fof.GroupLen           #number particles belonging to the group
        F_Mpart=F_mass[0]/F_part[0]   #mass of a single particle in Msun/h
        del fof

        #Correct the masses of the FoF halos
        F_mass=F_Mpart*(F_part*(1.0-F_part**(-0.6)))

        #some verbose
        print('Number of FoF halos=',len(F_pos))
        print(np.min(F_pos[:,0]),'< X_fof <',np.max(F_pos[:,0]))
        print(np.min(F_pos[:,1]),'< Y_fof <',np.max(F_pos[:,1]))
        print(np.min(F_pos[:,2]),'< Z_fof <',np.max(F_pos[:,2]))
        print(np.min(F_mass),'< M_fof <',np.max(F_mass))

        number=np.histogram(F_mass,bins=bins_mass)[0]
        print(number); print(np.sum(number,dtype=np.float64))

    elif obj=='halos_m200':
        #read CDM halos information
        halos=readsubf.subfind_catalog(groups_fname,groups_number,
                                       group_veldisp=True,masstab=True,
                                       long_ids=True,swap=False)

        H_pos=halos.group_pos/1e3               #positions in Mpc/h
        H_mass=halos.group_m_mean200*1e10       #masses in Msun/h
        H_radius_m=halos.group_r_mean200/1e3    #radius in Mpc/h
        del halos

        #some verbose
        print('Number of halos=',len(H_pos))
        print(np.min(H_pos[:,0]),'< X_fof <',np.max(H_pos[:,0]))
        print(np.min(H_pos[:,1]),'< Y_fof <',np.max(H_pos[:,1]))
        print(np.min(H_pos[:,2]),'< Z_fof <',np.max(H_pos[:,2]))
        print(np.min(H_mass),'< M_fof <',np.max(H_mass))

        number=np.histogram(H_mass,bins=bins_mass)[0]
        print(number); print(np.sum(number,dtype=np.float64))

    else:
        print('bad object type selected')
        sys.exit()

    sigma_mean=np.empty(bins); f_sigma=np.empty(bins)
    delta_f_sigma=np.empty(bins)
    for i in range(bins):
        R1=(3.0*bins_mass[i]/(4.0*pi*rhoM))**(1.0/3.0)
        sigma1=sigma(k,Pk,R1)

        R2=(3.0*bins_mass[i+1]/(4.0*pi*rhoM))**(1.0/3.0)
        sigma2=sigma(k,Pk,R2)

        sigma_mean[i]=0.5*(sigma2+sigma1)
        f_sigma[i]=(number[i]/np.log(sigma2/sigma1))/BoxSize**3
        f_sigma[i]=-(mass_mean[i]/rhoM)*f_sigma[i]
        delta_f_sigma[i]=(np.sqrt(number[i])/np.log(sigma2/sigma1))/BoxSize**3
        delta_f_sigma[i]=-(mass_mean[i]/rhoM)*delta_f_sigma[i]

    f=open(f_out,'w')
    for i in range(bins):
        f.write(str(sigma_mean[i])+' '+str(f_sigma[i])+' '+str(delta_f_sigma[i])+' '+str(mass_mean[i])+'\n')
    f.close()











import numpy as np
import sys

def mean(a):
    return np.sum(a)*1.0/len(a)

def standard_deviation(a):
    a_mean=np.sum(a)*1.0/len(a)
    b=np.sqrt(np.sum((a-a_mean)**2)/(len(a)-1.0))
    return b

def nearest(dist,boxsize):
    if (dist>+boxsize/2.0):
        dist-=boxsize
    if (dist<-boxsize/2.0):
        dist+=boxsize
    return dist

def pdf_log10(h,interv,minimum=0.0,maximum=0.0):
    if minimum==0.0 and maximum==0.0:
        minimo=np.min(h)
        maximo=np.max(h)
    else:
        minimo=minimum
        maximo=maximum
    bins_histo=np.logspace(np.log10(minimo),np.log10(maximo),interv+1)
    hist=np.histogram(h,bins=bins_histo)[0]
    if np.sum(hist)!=len(h) and minimum==0.0 and maximum==0.0:
        print('pdf does not contain all the elements')
    pdf=[]
    for i in range(interv):
        point1=bins_histo[i]
        point2=bins_histo[i+1]
        
        #choose one of the following definitions
        #pdf.append([10**(0.5*(np.log10(point1*point2))),hist[i]*1.0/(np.log(point2/point1)),np.sqrt(hist[i])/(np.log(point2/point1))])
        pdf.append([0.5*(point1+point2),hist[i]*1.0/(np.log(point2/point1)),np.sqrt(hist[i])/(np.log(point2/point1))])
        #pdf.append([10**(0.5*(np.log10(point1*point2))),hist[i]*1.0/(point2-point1),np.sqrt(hist[i])/(point2-point1)])                               

    pdf=np.array(pdf)
    return pdf

#pos is an array of 3 components representing the coordenate of the center of a halo. R is the radius of the sphere over which we want to compute. If that sphere, centered in the halo center intersects with box boundaries, returns True, otherwise return False
def in_border(pos,R,BoxSize):
    if any(pos+R>BoxSize) or any(pos-R<0):
        return True
    else:
        return False
################################################################################
#isolation_level determines the maximum mass of halos surrounding the main halo.
#isolation_level=0.1 means that a halo is considered isolated if there are no halos within times_radius x halos_radius[halo_index] with masses larger than 0.1 x halos_mass[halo_index]
def isolated(halo_index,halos_mass,halos_pos,halos_radius,BoxSize,times_radius,isolation_level=1.0):

    if isolation_level<1.0:
        indexes=np.where(halos_mass>halos_mass[halo_index]*isolation_level)[0]
        #remove the halo itself
        id=np.where(indexes==halo_index)[0]
        indexes=np.delete(indexes,id)
    else:
        indexes=np.where(halos_mass>halos_mass[halo_index]*isolation_level)[0]
    
    if indexes!=[]:
        distances=halos_pos[indexes]-halos_pos[halo_index]

        x=distances[:,0]
        y=distances[:,1]
        z=distances[:,2]
        del distances

        beyond=np.where(x>BoxSize/2.0)
        x[beyond]-=BoxSize
        beyond=np.where(x<-BoxSize/2.0)
        x[beyond]+=BoxSize

        beyond=np.where(y>BoxSize/2.0)
        y[beyond]-=BoxSize
        beyond=np.where(y<-BoxSize/2.0)
        y[beyond]+=BoxSize

        beyond=np.where(z>BoxSize/2.0)
        z[beyond]-=BoxSize
        beyond=np.where(z<-BoxSize/2.0)
        z[beyond]+=BoxSize

        distances=np.sqrt(x**2+y**2+z**2)
    
        if np.min(distances)<times_radius*halos_radius[halo_index]:
            return False
        else:
            return True
    else:
        return True
################################################################################
class PAR_inside:
    def __init__(self,PAR_pos,BoxSize,halo_pos,halo_radius,times_radius):

        R=halo_radius*times_radius

        PAR_X=PAR_pos[:,0]-halo_pos[0]
        PAR_Y=PAR_pos[:,1]-halo_pos[1]
        PAR_Z=PAR_pos[:,2]-halo_pos[2]
        del PAR_pos

        if any(halo_pos+R>BoxSize) or any(halo_pos-R<0.0):
            beyond=np.where(PAR_X>BoxSize/2.0)
            PAR_X[beyond]-=BoxSize
            beyond=np.where(PAR_X<-BoxSize/2.0)
            PAR_X[beyond]+=BoxSize

            beyond=np.where(PAR_Y>BoxSize/2.0)
            PAR_Y[beyond]-=BoxSize
            beyond=np.where(PAR_Y<-BoxSize/2.0)
            PAR_Y[beyond]+=BoxSize

            beyond=np.where(PAR_Z>BoxSize/2.0)
            PAR_Z[beyond]-=BoxSize
            beyond=np.where(PAR_Z<-BoxSize/2.0)
            PAR_Z[beyond]+=BoxSize

        radius=np.sqrt(PAR_X**2+PAR_Y**2+PAR_Z**2)
        del PAR_X,PAR_Y,PAR_Z
        inside=np.where(radius<R)[0]
        self.len=len(inside)
        self.id=inside
        del radius,inside

class DM_inside:
    def __init__(self,DM_pos,BoxSize,rhocrit,Omega_DM,halo_pos,halo_virial_radius,id_number,times_virial_radius):
        self.id=id_number

        R=halo_virial_radius*times_virial_radius

        DM_X=DM_pos[:,0]-halo_pos[0]
        DM_Y=DM_pos[:,1]-halo_pos[1]
        DM_Z=DM_pos[:,2]-halo_pos[2]

        if any(halo_pos+R>BoxSize) or any(halo_pos-R<0.0):
            beyond=np.where(DM_X>BoxSize/2.0)
            DM_X[beyond]-=BoxSize
            beyond=np.where(DM_X<-BoxSize/2.0)
            DM_X[beyond]+=BoxSize

            beyond=np.where(DM_Y>BoxSize/2.0)
            DM_Y[beyond]-=BoxSize
            beyond=np.where(DM_Y<-BoxSize/2.0)
            DM_Y[beyond]+=BoxSize

            beyond=np.where(DM_Z>BoxSize/2.0)
            DM_Z[beyond]-=BoxSize
            beyond=np.where(DM_Z<-BoxSize/2.0)
            DM_Z[beyond]+=BoxSize

        radius=np.sqrt(DM_X**2+DM_Y**2+DM_Z**2)
        del DM_X,DM_Y,DM_Z
        inside=np.where(radius<R)[0]
        self.DM_len=len(inside)

        f=open('DMhalo'+str(id_number),'w')
        f.write (str(halo_pos[0])+' '+str(halo_pos[1])+' '+str(halo_pos[2])+' '+str(halo_virial_radius)+'\n')
        for i in range(len(inside)):
            f.write(str(DM_pos[inside[i],0])+' '+str(DM_pos[inside[i],1])+' '+str(DM_pos[inside[i],2])+'\n')
        f.close()
        del radius,inside


class PAR_cylindrical:
    #note that intervals has to be an odd number. When calling the function it is a good practice to do 2*intervals+1, where intervals are the number of bins over which the profile is computed
    def __init__(self,PAR_pos,BoxSize,halo_pos,halo_radius,h_max,V,intervals):

        PAR_pos=PAR_pos-halo_pos

        Rmax=np.sqrt(halo_radius**2+h_max**2) #max distance to halo center
        if any(halo_pos+Rmax>BoxSize) or any(halo_pos-Rmax<0.0):
            beyond=np.where(PAR_pos[:,0]>BoxSize/2.0)
            PAR_pos[beyond,0]-=BoxSize
            beyond=np.where(PAR_pos[:,0]<-BoxSize/2.0)
            PAR_pos[beyond,0]+=BoxSize

            beyond=np.where(PAR_pos[:,1]>BoxSize/2.0)
            PAR_pos[beyond,1]-=BoxSize
            beyond=np.where(PAR_pos[:,1]<-BoxSize/2.0)
            PAR_pos[beyond,1]+=BoxSize

            beyond=np.where(PAR_pos[:,2]>BoxSize/2.0)
            PAR_pos[beyond,2]-=BoxSize
            beyond=np.where(PAR_pos[:,2]<-BoxSize/2.0)
            PAR_pos[beyond,2]+=BoxSize

        #compute the minimum distance between a point and the axis V
        numerator=np.cross(PAR_pos,V)
        numerator=numerator[:,0]**2+numerator[:,1]**2+numerator[:,2]**2
        radius_h=np.sqrt(numerator/np.dot(V,V))
        del numerator
        #compute the distance between halo center and the projection of a point over the V axis
        z=np.dot(PAR_pos,V)/np.sqrt(np.dot(V,V))

        inside1=radius_h<halo_radius
        inside2=np.abs(z)<h_max
        inside=np.where(inside1*inside2==True)
        particles=PAR_pos[inside]
        z=z[inside]
        del PAR_pos,radius_h

        interv=np.linspace(-h_max,h_max,intervals)
        hist=np.histogram(z,bins=interv)[0] #the histogram is created over intervals-1 bins
        if np.sum(hist).astype(np.int64)!=len(particles):
            print('not all particles sorted')
            print(np.sum(hist),len(particles))
            sys.exit()
        del z

        length=int((intervals-1)//2)
        bin=np.empty(length,dtype=np.float32)
        values=np.empty(length,dtype=np.float32)
        errors=np.empty(length,dtype=np.float32)
        for i in range(length):
            bin[i]=abs(0.5*(interv[i]+interv[i+1]))/halo_radius
            if hist[intervals-2-i]==0 and hist[i]==0:
                values[i]=0.0
                errors[i]=0.0
            else:
                a=hist[intervals-2-i]*1.0
                b=hist[i]*1.0
                values[i]=(a-b)/(a+b)
                errors[i]=np.sqrt((1.0+values[i]**2)/(a+b))
           
        self.PAR_len=len(inside)
        self.id=inside
        self.bins=bin
        self.values=values
        self.errors=errors





    



################################################################################
class PAR_profile:
    def __init__(self,PAR_pos,PAR_nall,BoxSize,halo_index,halo_pos,halo_radius,times_radius,number_of_bins,radius_first_bin,r_bin_max,mean_R,write_file=True):

        #R=halo_radius*times_radius
        R=mean_R

        PAR_X=PAR_pos[:,0]-halo_pos[0]
        PAR_Y=PAR_pos[:,1]-halo_pos[1]
        PAR_Z=PAR_pos[:,2]-halo_pos[2]
        del PAR_pos

        if any(halo_pos+R>BoxSize) or any(halo_pos-R<0.0):
            beyond=np.where(PAR_X>BoxSize/2.0)
            PAR_X[beyond]-=BoxSize
            beyond=np.where(PAR_X<-BoxSize/2.0)
            PAR_X[beyond]+=BoxSize

            beyond=np.where(PAR_Y>BoxSize/2.0)
            PAR_Y[beyond]-=BoxSize
            beyond=np.where(PAR_Y<-BoxSize/2.0)
            PAR_Y[beyond]+=BoxSize

            beyond=np.where(PAR_Z>BoxSize/2.0)
            PAR_Z[beyond]-=BoxSize
            beyond=np.where(PAR_Z<-BoxSize/2.0)
            PAR_Z[beyond]+=BoxSize

        radius=np.sqrt(PAR_X**2+PAR_Y**2+PAR_Z**2)
        del PAR_X,PAR_Y,PAR_Z
        inside=np.where(radius<R)[0]
        radius_R=radius[inside]
        del inside,radius

        max_radius=np.max(radius_R)
        if (max_radius>R):
            print('error',max_radius)
            sys.exit()

        PAR_background_density=PAR_nall/BoxSize**3

        bin=np.zeros(number_of_bins,dtype=np.uint32)
        overdensity=np.empty(number_of_bins)
        #r_bin_max=np.logspace(np.log10(radius_first_bin),np.log10(R),number_of_bins)
        r_bin_min=np.empty(number_of_bins)
        r_bin_min[0]=0.0
        r_bin_min[1:number_of_bins]=r_bin_max[0:number_of_bins-1]
        r_bin=0.5*(r_bin_min+r_bin_max)
        volume_bin=4.0*np.pi*(r_bin_max**3-r_bin_min**3)/3.0
        for j in range(number_of_bins):
            len_bin=len(np.where(radius_R<=r_bin_max[j])[0])
            bin[j]=len_bin-np.sum(bin)
            overdensity[j]=bin[j]*1.0/(volume_bin[j]*PAR_background_density)

        if np.sum(bin).astype(np.uint32)!=len(radius_R):
            print('not all particles counted')
            print('counted ',np.sum(bin),' of a total of',len(radius))
            sys.exit()
        del radius_R,r_bin_max,r_bin_min,bin

        self.bin_radius=r_bin
        self.bin_overdensity=overdensity

        if write_file:
            f=open('NUprofile'+str(halo_index),'w')
            for j in range(number_of_bins):
                f.write(str(r_bin[j])+' '+str(overdensity[j])+'\n')
            f.close()
################################################################################
#compute the velocity distribution within a given CDM halo
class VEL_profile_halo:
    def __init__(self,PAR_pos,PAR_vel,BoxSize,halo_pos,halo_radius,times_radius,interv,V_min=0.0,V_max=0.0,scale='lin'):

        R=halo_radius*times_radius

        PAR_X=PAR_pos[:,0]-halo_pos[0]
        PAR_Y=PAR_pos[:,1]-halo_pos[1]
        PAR_Z=PAR_pos[:,2]-halo_pos[2]
        del PAR_pos

        if any(halo_pos+R>BoxSize) or any(halo_pos-R<0.0):
            beyond=np.where(PAR_X>BoxSize/2.0)
            PAR_X[beyond]-=BoxSize
            beyond=np.where(PAR_X<-BoxSize/2.0)
            PAR_X[beyond]+=BoxSize

            beyond=np.where(PAR_Y>BoxSize/2.0)
            PAR_Y[beyond]-=BoxSize
            beyond=np.where(PAR_Y<-BoxSize/2.0)
            PAR_Y[beyond]+=BoxSize

            beyond=np.where(PAR_Z>BoxSize/2.0)
            PAR_Z[beyond]-=BoxSize
            beyond=np.where(PAR_Z<-BoxSize/2.0)
            PAR_Z[beyond]+=BoxSize

        radius=np.sqrt(PAR_X**2+PAR_Y**2+PAR_Z**2)
        del PAR_X,PAR_Y,PAR_Z
        inside=np.where(radius<R)[0]

        V=np.sqrt(PAR_vel[inside,0]**2+PAR_vel[inside,1]**2+PAR_vel[inside,2]**2)
        del inside
        if V_min==0.0 and V_max==0.0:
            V_min=np.min(V)
            V_max=np.max(V)
        if scale=='lin':
            bins_histo=np.linspace(V_min,V_max,interv+1)
        elif scale=='log':
            bins_histo=np.logspace(np.log10(V_min),np.log10(V_max),interv+1)
        else:
            print('error:choose scale lin or log')
            sys.exit()

        if len(V)!=0:
            hist=np.histogram(V,bins=bins_histo)[0]*1.0/len(V)
            self.bin_h=hist
        else:
            self.bin_h=np.zeros(interv,dtype=np.float64)

        self.bin_v=bins_histo


def bulk_V(PAR_pos,PAR_vel,halo_pos,halo_radius,times_radius,BoxSize,rms=False):

    R=halo_radius*times_radius
    PAR_pos=PAR_pos-halo_pos

    #this is valid as long as R is much smaller than the BoxSize
    if any(halo_pos+R>BoxSize) or any(halo_pos-R<0.0):
        beyond=np.where(PAR_pos[:,0]>BoxSize/2.0)
        PAR_pos[beyond,0]-=BoxSize
        beyond=np.where(PAR_pos[:,0]<-BoxSize/2.0)
        PAR_pos[beyond,0]+=BoxSize

        beyond=np.where(PAR_pos[:,1]>BoxSize/2.0)
        PAR_pos[beyond,1]-=BoxSize
        beyond=np.where(PAR_pos[:,1]<-BoxSize/2.0)
        PAR_pos[beyond,1]+=BoxSize
        
        beyond=np.where(PAR_pos[:,2]>BoxSize/2.0)
        PAR_pos[beyond,2]-=BoxSize
        beyond=np.where(PAR_pos[:,2]<-BoxSize/2.0)
        PAR_pos[beyond,2]+=BoxSize
        del beyond

    radius=np.sqrt(PAR_pos[:,0]**2+PAR_pos[:,1]**2+PAR_pos[:,2]**2)
    del PAR_pos
    inside=np.where(radius<R)

    number_of_particles=len(inside[0])*1.0

    print('number of particles=',number_of_particles)

    max_radius=np.max(radius[inside])
    if (max_radius>R):
        print('error',max_radius)
        sys.exit()
    del radius

    bulk_vel=np.zeros(3,np.float32)
    bulk_vel[0]=np.sum(PAR_vel[inside[0],0])
    bulk_vel[1]=np.sum(PAR_vel[inside[0],1])
    bulk_vel[2]=np.sum(PAR_vel[inside[0],2])
    bulk_vel_mod=np.sqrt(bulk_vel[0]**2+bulk_vel[1]**2+bulk_vel[2]**2)
    
    velocities=PAR_vel[inside[0],0]**2+PAR_vel[inside[0],1]**2+PAR_vel[inside[0],2]**2
    del inside
    rms=np.sqrt(np.sum(velocities)/len(velocities))
    V_mean=np.sum(np.sqrt(velocities))/len(velocities)
    sigma=np.sqrt(np.sum((np.sqrt(velocities)-V_mean)**2)/len(velocities))

    return bulk_vel/number_of_particles,V_mean,rms,sigma,number_of_particles
###############################################################################
#compute the peculiar velocity of particles in shell within radius
#halo_radius x times_radius1 and halo_radius x times_radius2
#times_radius2 has to be larger than times_radius1
def bulk_shell_V(PAR_pos,PAR_vel,halo_pos,halo_radius,times_radius1,times_radius2,BoxSize):

    R2=halo_radius*times_radius2
    R1=halo_radius*times_radius1
    PAR_pos=PAR_pos-halo_pos

    #this is valid as long as R is much smaller than the BoxSize
    if any(halo_pos+R2>BoxSize) or any(halo_pos-R2<0.0):
        beyond=np.where(PAR_pos[:,0]>BoxSize/2.0)
        PAR_pos[beyond,0]-=BoxSize
        beyond=np.where(PAR_pos[:,0]<-BoxSize/2.0)
        PAR_pos[beyond,0]+=BoxSize

        beyond=np.where(PAR_pos[:,1]>BoxSize/2.0)
        PAR_pos[beyond,1]-=BoxSize
        beyond=np.where(PAR_pos[:,1]<-BoxSize/2.0)
        PAR_pos[beyond,1]+=BoxSize
        
        beyond=np.where(PAR_pos[:,2]>BoxSize/2.0)
        PAR_pos[beyond,2]-=BoxSize
        beyond=np.where(PAR_pos[:,2]<-BoxSize/2.0)
        PAR_pos[beyond,2]+=BoxSize
        del beyond

    radius=np.sqrt(PAR_pos[:,0]**2+PAR_pos[:,1]**2+PAR_pos[:,2]**2)
    del PAR_pos
    inside1=np.where(radius<R1)[0]
    inside2=np.where(radius<R2)[0]
    n_particles=len(inside2)-len(inside1)

    bulk_vel=np.zeros(3,np.float32)
    bulk_vel[0]=np.sum(PAR_vel[inside2,0])-np.sum(PAR_vel[inside1,0])
    bulk_vel[1]=np.sum(PAR_vel[inside2,1])-np.sum(PAR_vel[inside1,1])
    bulk_vel[2]=np.sum(PAR_vel[inside2,2])-np.sum(PAR_vel[inside1,2])
    del inside1,inside2

    return bulk_vel/n_particles
###############################################################################
def vel_dispersion(PAR_pos,PAR_vel,halo_pos,halo_radius,times_radius,BoxSize):

    R=halo_radius*times_radius
    PAR_pos=PAR_pos-halo_pos

    #this is valid as long as R is much smaller than the BoxSize
    if any(halo_pos+R>BoxSize) or any(halo_pos-R<0.0):
        beyond=np.where(PAR_pos[:,0]>BoxSize/2.0)
        PAR_pos[beyond,0]-=BoxSize
        beyond=np.where(PAR_pos[:,0]<-BoxSize/2.0)
        PAR_pos[beyond,0]+=BoxSize

        beyond=np.where(PAR_pos[:,1]>BoxSize/2.0)
        PAR_pos[beyond,1]-=BoxSize
        beyond=np.where(PAR_pos[:,1]<-BoxSize/2.0)
        PAR_pos[beyond,1]+=BoxSize
        
        beyond=np.where(PAR_pos[:,2]>BoxSize/2.0)
        PAR_pos[beyond,2]-=BoxSize
        beyond=np.where(PAR_pos[:,2]<-BoxSize/2.0)
        PAR_pos[beyond,2]+=BoxSize
        del beyond

    radius=np.sqrt(PAR_pos[:,0]**2+PAR_pos[:,1]**2+PAR_pos[:,2]**2)
    del PAR_pos
    inside=np.where(radius<R)[0]

    number_of_particles=len(inside)*1.0

    max_radius=np.max(radius[inside])
    if (max_radius>R):
        print('error',max_radius)
        sys.exit()
    del radius

    bulk_vel=np.zeros(3,np.float32)
    bulk_vel[0]=np.sum(PAR_vel[inside,0])
    bulk_vel[1]=np.sum(PAR_vel[inside,1])
    bulk_vel[2]=np.sum(PAR_vel[inside,2])
    bulk_vel=bulk_vel/number_of_particles

    vel=PAR_vel[inside]-bulk_vel
    dispersion=np.sum(vel[:,0]**2+vel[:,1]**2+vel[:,2]**2)/(3.0*len(vel))
    dispersion=np.sqrt(dispersion)

    return dispersion
###############################################################################
def low_V(PAR_pos,PAR_vel,halo_pos,halo_radius,times_radius,BoxSize):

    R=halo_radius*times_radius
    PAR_pos=PAR_pos-halo_pos

    #this is valid as long as R is much smaller than the BoxSize
    if any(halo_pos+R>BoxSize) or any(halo_pos-R<0.0):
        beyond=np.where(PAR_pos[:,0]>BoxSize/2.0)
        PAR_pos[beyond,0]-=BoxSize
        beyond=np.where(PAR_pos[:,0]<-BoxSize/2.0)
        PAR_pos[beyond,0]+=BoxSize

        beyond=np.where(PAR_pos[:,1]>BoxSize/2.0)
        PAR_pos[beyond,1]-=BoxSize
        beyond=np.where(PAR_pos[:,1]<-BoxSize/2.0)
        PAR_pos[beyond,1]+=BoxSize
        
        beyond=np.where(PAR_pos[:,2]>BoxSize/2.0)
        PAR_pos[beyond,2]-=BoxSize
        beyond=np.where(PAR_pos[:,2]<-BoxSize/2.0)
        PAR_pos[beyond,2]+=BoxSize
        del beyond

    radius=np.sqrt(PAR_pos[:,0]**2+PAR_pos[:,1]**2+PAR_pos[:,2]**2)
    del PAR_pos
    inside=np.where(radius<R)[0]

    number_of_particles=len(inside)
    print('number of particles=',number_of_particles)

    max_radius=np.max(radius[inside])
    if (max_radius>R):
        print('error',max_radius)
        sys.exit()
    del radius

    vel=PAR_vel[inside]
    mod_vel=np.sqrt(vel[:,0]**2+vel[:,1]**2+vel[:,2]**2)
    low=np.where(mod_vel<400.0)[0]

    print(low)
    return inside,low
    


################################################################################
#this routine returns the position of the center of mass of a given particle distribution
def COM(PAR_pos,halo_pos,halo_radius,times_radius,BoxSize):

    R=halo_radius*times_radius
    PAR_pos=PAR_pos-halo_pos

    #this is valid as long as R is much smaller than the BoxSize
    if any(halo_pos+R>BoxSize) or any(halo_pos-R<0.0):
        beyond=np.where(PAR_pos[:,0]>BoxSize/2.0)
        PAR_pos[beyond,0]-=BoxSize
        beyond=np.where(PAR_pos[:,0]<-BoxSize/2.0)
        PAR_pos[beyond,0]+=BoxSize

        beyond=np.where(PAR_pos[:,1]>BoxSize/2.0)
        PAR_pos[beyond,1]-=BoxSize
        beyond=np.where(PAR_pos[:,1]<-BoxSize/2.0)
        PAR_pos[beyond,1]+=BoxSize
        
        beyond=np.where(PAR_pos[:,2]>BoxSize/2.0)
        PAR_pos[beyond,2]-=BoxSize
        beyond=np.where(PAR_pos[:,2]<-BoxSize/2.0)
        PAR_pos[beyond,2]+=BoxSize
        del beyond

    radius=np.sqrt(PAR_pos[:,0]**2+PAR_pos[:,1]**2+PAR_pos[:,2]**2)
    inside=np.where(radius<R)

    max_radius=np.max(radius[inside])
    if (max_radius>R):
        print('error',max_radius)
        sys.exit()
    del radius

    COM=np.zeros(3,np.float64)
    COM[0]=np.sum(PAR_pos[inside[0],0],dtype=np.float64)
    COM[1]=np.sum(PAR_pos[inside[0],1],dtype=np.float64)
    COM[2]=np.sum(PAR_pos[inside[0],2],dtype=np.float64)
    del PAR_pos

    return COM/len(inside[0])
###############################################################################

################################################################################
#this routine returns the number of particles within a given radius from a point
def interior_mass(PAR_pos,halo_pos,halo_radius,times_radius,BoxSize):

    R=halo_radius*times_radius

    PAR_X=PAR_pos[:,0]-halo_pos[0]
    PAR_Y=PAR_pos[:,1]-halo_pos[1]
    PAR_Z=PAR_pos[:,2]-halo_pos[2]
    del PAR_pos

    #this is valid as long as R is much smaller than the BoxSize
    if any(halo_pos+R>BoxSize) or any(halo_pos-R<0.0):
        beyond=np.where(PAR_X>BoxSize/2.0)
        PAR_X[beyond]-=BoxSize
        beyond=np.where(PAR_X<-BoxSize/2.0)
        PAR_X[beyond]+=BoxSize

        beyond=np.where(PAR_Y>BoxSize/2.0)
        PAR_Y[beyond]-=BoxSize
        beyond=np.where(PAR_Y<-BoxSize/2.0)
        PAR_Y[beyond]+=BoxSize
        
        beyond=np.where(PAR_Z>BoxSize/2.0)
        PAR_Z[beyond]-=BoxSize
        beyond=np.where(PAR_Z<-BoxSize/2.0)
        PAR_Z[beyond]+=BoxSize
        del beyond

    radius=np.sqrt(PAR_X**2+PAR_Y**2+PAR_Z**2)
    del PAR_X,PAR_Y,PAR_Z
    inside=np.where(radius<=R)
    number_of_particles=len(inside[0])*1.0

    if number_of_particles>0.0:
        max_radius=np.max(radius[inside])
        if (max_radius>R):
            print('error',max_radius)
            sys.exit()
    del radius

    return number_of_particles
















import numpy as np
import sys,os,time
import Pk_library as PKL
import units_library as UL
cimport numpy as np
cimport cython
from cython.parallel import prange,parallel
from libc.math cimport sqrt,pow,sin,cos,log,log10,fabs,round
from libc.stdlib cimport malloc, free
from libc.stdio cimport printf
cimport void_openmp_library as VOL

DEF PI=3.141592653589793

############################### ROUTINES ####################################
# V = void_finder(delta, BoxSize, threshold, threads, void_field=False)
# V.void_pos  ----> void positions
# V.void_radius --> void radii  
# V.Rbins --------> void mass function (R bins)
# V.void_vsf -----> void mass function (voids/(Volume*log(R)))
# V.in_void ------> grid with 0 and 1. 1 is a cell within a void (optional)

# gaussian_smoothing(delta, BoxSize, R, threads)

# V = void_safety_check(delta, void_pos, void_radius, BoxSize)
# V.mean_overdensity ----> mean overdensity of the cells in a void
# V.mean_radius ---------> effective radius of the void from the cells in it

# V = random_spheres(BoxSize, Rmin, Rmax, Nvoids, dims)
# V.void_pos ------> array with the positions of the input voids
# V.void_radius ---> array with the radii of the input void
# V.delta ---------> density field with zeros everywhere except in voids
#############################################################################


# This function sorts the input Radii, from largest to smallest
def sort_Radii(float[:] Radii):
    return np.sort(Radii)[::-1]

# The function takes a density field and smooth it with a 3D top-hat filter
# of radius R:  W = 3/(4*pi*R^3) if r<R;  W = 0  otherwise
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
def gaussian_smoothing(delta, float BoxSize, float R, int threads=1):
                       
    cdef int dims = delta.shape[0]
    cdef int middle = dims//2
    cdef float prefact,kR,fact
    cdef int kxx, kyy, kzz, kx, ky, kz, kx2, ky2, kz2
    cdef np.complex64_t[:,:,::1] delta_k

    ## compute FFT of the field (change this for double precision) ##
    delta_k = PKL.FFT3Dr_f(delta,threads) 

    # do a loop over the independent modes.
    prefact = R*2.0*PI/BoxSize
    for kxx in prange(dims, nogil=True):
        kx  = (kxx-dims if (kxx>middle) else kxx)
        kx2 = kx*kx
        
        for kyy in range(dims):
            ky  = (kyy-dims if (kyy>middle) else kyy)
            ky2 = ky*ky
            
            for kzz in range(middle+1): #kzz=[0,1,..,middle] --> kz>0
                kz  = (kzz-dims if (kzz>middle) else kzz)
                kz2 = kz*kz

                if kxx==0 and kyy==0 and kzz==0:
                    continue

                # compute the value of |k|
                kR = prefact*sqrt(kx2 + ky2 + kz2)
                if fabs(kR)<1e-5:  fact = 1.0
                else:              fact = 3.0*(sin(kR) - cos(kR)*kR)/(kR*kR*kR)
                delta_k[kxx,kyy,kzz] = delta_k[kxx,kyy,kzz]*fact
                                       
    # Fourier transform back
    return PKL.IFFT3Dr_f(delta_k,threads)

# This routine finds voids in the density field delta. The input values needed
# are the box size, the maximum and minimum radii, the number of bins, the
# density threshold, Omega_m (to compute void masses) and the threads number
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
class void_finder:
    def __init__(self, np.ndarray[np.float32_t, ndim=3] delta, float BoxSize, 
        float threshold, float[:] Radii,
        int threads, int threads2, void_field=False):

        cdef float R, dist2, R_grid, R_grid2, Rmin
        cdef float dx, dy, dz, middle
        cdef long voids_found, total_voids_found,num
        cdef long max_num_voids,local_voids,ID,dims3
        cdef int i,j,k,p,q,Ncells,l,m,n,i1,j1,k1, nearby_voids
        cdef int dims, dims2, mode, bins
        cdef char[:,:,::1] in_void
        cdef float[::1] delta_v, delta_v_temp
        cdef long[::1] IDs, indexes, IDs_temp
        cdef float[:] vsf, Rmean
        cdef int[::1] Nvoids
        cdef float[:,:,::1] delta_sm
        cdef int[:,::1] void_pos
        cdef float[::1] void_mass
        cdef float[::1] void_radius
        cdef double expected_filling_factor=0.0
        cdef double time1,time2,dt,time_tot

        time_tot = time.time()
        dims = delta.shape[0];  middle = dims//2
        dims2 = dims**2;  dims3 = dims**3
        bins = Radii.shape[0]

        # sort the input Radii
        Radii = sort_Radii(Radii)

        # check that Rmin is larger than the grid resolution
        Rmin = np.min(Radii)
        if Rmin<BoxSize*1.0/dims:
            raise Exception("Rmin=%.3f Mpc/h below grid resolution=%.3f Mpc/h"\
                            %(Rmin, BoxSize*1.0/dims))

        # find the maximum possible number of voids
        max_num_voids = int(BoxSize**3/(4.0*PI/3.0*Rmin**3))
        print('maximum number of voids = %d\n'%max_num_voids)

        # define arrays containing void positions and radii
        void_pos    = np.zeros((max_num_voids, 3), dtype=np.int32)
        void_radius = np.zeros(max_num_voids,      dtype=np.float32)
        
        # define the in_void and delta_v array
        in_void = np.zeros((dims,dims,dims), dtype=np.int8)
        delta_v = np.zeros(dims3,            dtype=np.float32)
        IDs     = np.zeros(dims3,            dtype=np.int64)

        # define the arrays needed to compute the VSF
        Nvoids = np.zeros(bins,   dtype=np.int32)
        vsf    = np.zeros(bins-1, dtype=np.float32)
        Rmean  = np.zeros(bins-1, dtype=np.float32)

        total_voids_found = 0
        for q in range(bins):

            # Get the smoothing length
            R = Radii[q]

            # smooth the density field with a top-hat radius of R
            start = time.time()
            print('Smoothing field with top-hat filter of radius %.2f'%R)
            delta_sm = gaussian_smoothing(delta, BoxSize, R, threads)
            print('Density smoothing took %.3f seconds'%(time.time()-start))
            if np.min(delta_sm)>threshold:
                print('No cells with delta < %.2f\n'%threshold)
                continue

            ######### find underdense cells ##########
            # cells with delta<threshold and not in existing voids
            start = time.time()
            local_voids = 0
            for i in range(dims):
                for j in range(dims):
                    for k in range(dims):

                        if delta_sm[i,j,k]<threshold and in_void[i,j,k]==0:
                            IDs[local_voids]     = dims2*i + dims*j + k
                            delta_v[local_voids] = delta_sm[i,j,k]
                            local_voids += 1
            print('Searching underdense cells took %.3f seconds'%(time.time()-start))
            print('Found %08d cells below threshold'%(local_voids))
            ##########################################

            ######## sort cell underdensities ########
            # sort cells by their underdensity
            start = time.time()
            indexes = np.argsort(delta_v[:local_voids])

            # this is just delta_v = delta_v[indexes]
            delta_v_temp = np.empty(local_voids, dtype=np.float32)
            for i in range(local_voids):
                delta_v_temp[i] = delta_v[indexes[i]]
            for i in range(local_voids):
                delta_v[i] = delta_v_temp[i]
            del delta_v_temp

            # this is just IDs = IDs[indexes]
            IDs_temp = np.empty(local_voids, dtype=np.int64) 
            for i in range(local_voids):
                IDs_temp[i] = IDs[indexes[i]]
            for i in range(local_voids):
                IDs[i] = IDs_temp[i]
            del IDs_temp

            print('Sorting took %.3f seconds'%(time.time()-start))
            ##########################################

            # do a loop over all underdense cells and identify voids
            start   = time.time()
            R_grid  = (R/BoxSize)*1.0*dims;  Ncells = <int>R_grid + 1
            R_grid2 = R_grid*R_grid
            voids_found = 0

            if total_voids_found<(2*Ncells+1)**3:  mode = 0
            else:                                  mode = 1
            time1, time2 = 0.0, 0.0
            if Ncells<12:  threads2 = 1 #empirically this seems to be the best
            print('Mode = %d    :   Ncells = %d   :   threads = %d'%(mode,Ncells,threads2))
            for p in range(local_voids):

                # find the grid coordinates of the underdense cell
                ID = IDs[p]
                i,j,k = ID//dims2, (ID%dims2)//dims, (ID%dims2)%dims

                # if cell belong to a void continue
                if in_void[i,j,k] == 1:  continue

                # find if there are voids overlapping with this void candidate either
                # by computing distances to other voids (mode=0) or searching for
                # in_void=1 in cells belonging to void canditate (mode=1)
                dt = time.time()
                if mode==0:
                    nearby_voids = VOL.num_voids_around(total_voids_found, dims, 
                                                        middle, i, j, k, 
                                                        &void_radius[0], 
                                                        &void_pos[0,0], R_grid, 
                                                        threads2)
                else:
                    nearby_voids = VOL.num_voids_around2(Ncells, i, j, k, dims,
                                                         R_grid2, &in_void[0,0,0], 
                                                         threads2)
                time1 += (time.time()-dt)
                """ #old num_voids_around routine
                nearby_voids = 0
                for l in prange(total_voids_found, nogil=True):

                    dx = i-void_pos[l,0]
                    if dx>middle:   dx = dx - dims
                    if dx<-middle:  dx = dx + dims

                    dy = j-void_pos[l,1]
                    if dy>middle:   dy = dy - dims
                    if dy<-middle:  dy = dy + dims

                    dz = k-void_pos[l,2]
                    if dz>middle:   dz = dz - dims
                    if dz<-middle:  dz = dz + dims

                    dist2 = dx*dx + dy*dy + dz*dz

                    if dist2<(void_radius[l]+R_grid)*(void_radius[l]+R_grid):
                        nearby_voids += 1
                        break
                """
                
                """ #old num_voids_around2 routine
                # check that all cells in the void are not in other void
                cells_in_other_void = 0
                for l in prange(-Ncells,Ncells+1, nogil=True):
                     i1 = (i+l+dims)%dims

                     for m in range(-Ncells,Ncells+1):
                         j1 = (j+m+dims)%dims
                        
                         for n in range(-Ncells,Ncells+1):
                             k1 = (k+n+dims)%dims
                                
                             dist2 = l*l + m*m + n*n
                             if dist2<R_grid2 and in_void[i1,j1,k1]==1:
                                 cells_in_other_void += 1
                                 break
                """

                # we have found a new void
                if nearby_voids==0:

                    void_pos[total_voids_found, 0] = i
                    void_pos[total_voids_found, 1] = j
                    void_pos[total_voids_found, 2] = k
                    void_radius[total_voids_found] = R_grid

                    voids_found += 1;  total_voids_found += 1 
                    in_void[i,j,k] = 1
                    
                    # put in_void[i,j,k]=1 to the cells belonging to the void
                    # it seems that with 1 thread is more than enough
                    dt = time.time()
                    VOL.mark_void_region(&in_void[0,0,0], Ncells, dims, R_grid2,
                                         i, j, k, threads=1)
                    time2 += (time.time()-dt)


            print('Found %06d voids with radius R =%.3f Mpc/h'%(voids_found, R))
            print('Found %06d voids with radius R>=%.3f Mpc/h'%(total_voids_found,R))
            print('Void volume filling fraction = %.3e'\
                %(np.sum(in_void, dtype=np.int64)*1.0/dims3))
            expected_filling_factor += voids_found*4.0*np.pi/3.0*R**3/BoxSize**3
            print('Expected    filling fraction = %.3e'%(expected_filling_factor))
            print('Time1 = %.3f seconds'%time1)
            print('Time2 = %.3f seconds'%time2)
            print('void finding took %.3f seconds\n'%(time.time()-start))
            Nvoids[q] = voids_found   

        print('Void volume filling fraction = %.3f'\
            %(np.sum(in_void, dtype=np.int64)*1.0/dims3))
        print('Found a total of %d voids'%total_voids_found)
        print('Total time take %.3f seconds\n'%(time.time()-time_tot))

        # compute the void size function (# of voids/Volume/dR)
        for i in range(bins-1):
            vsf[i]   = Nvoids[i]/(BoxSize**3*(Radii[i]-Radii[i+1]))
            Rmean[i] = 0.5*(Radii[i]+Radii[i+1])

        # finish by setting the class fields
        self.void_pos    = np.asarray(void_pos[:total_voids_found])*(BoxSize/dims)
        self.void_radius = np.asarray(void_radius[:total_voids_found])*(BoxSize/dims)
        self.Rbins       = np.asarray(Rmean)
        self.void_vsf    = np.asarray(vsf)
        if void_field:   self.in_void = np.asarray(in_void)

# This routine takes the density field, the void positions and radii
# and finds the cells inside each void. From those cells it computes
# the mean overdensity, the effective radius and the mass of the void
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
class void_safety_check:
    def __init__(self, float[:,:,:] delta, float[:,:] void_pos, 
        float[:] void_radius, float BoxSize):

        cdef long number_of_voids, p
        cdef int dims, i, j, k, Nshells, ii, jj, kk, i1, j1, k1
        cdef float dist, R_void, prefact, ratio, V_cell
        cdef double[:] mean_overdensity, mean_radius
        cdef long[:] cells

        # find the number of voids and the number of cells in the field
        number_of_voids = void_pos.shape[0]
        dims            = delta.shape[0]

        # define the array containing the mean overdensities
        mean_overdensity = np.zeros(number_of_voids, dtype=np.float64)
        mean_radius      = np.zeros(number_of_voids, dtype=np.float64)
        cells            = np.zeros(number_of_voids, dtype=np.int64)

        prefact = dims*1.0/BoxSize
        V_cell  = (BoxSize*1.0/dims)**3

        # do a loop over all the voids
        for p in range(number_of_voids):

            # find void coordinate in grid units
            i = <int>round(void_pos[p,0]*prefact)
            j = <int>round(void_pos[p,1]*prefact)
            k = <int>round(void_pos[p,2]*prefact)

            # compute the void radius in grid cell units
            R_void  = void_radius[p]*prefact
            Nshells = <int>R_void+1

            for ii in range(-Nshells, Nshells+1):
                i1 = (ii+i+dims)%dims
                for jj in range(-Nshells, Nshells+1):
                    j1 = (jj+j+dims)%dims
                    for kk in range(-Nshells, Nshells+1):
                        k1 = (kk+k+dims)%dims

                        dist = sqrt(ii*ii + jj*jj + kk*kk)
                        if dist<R_void:
                            mean_overdensity[p] += delta[i1,j1,k1]
                            cells[p] += 1

            # compute the mean overdensity of the cells in the void
            mean_overdensity[p] = mean_overdensity[p]*1.0/cells[p]

            # compute volume occupied by cells in void and effective radius
            mean_radius[p] = (3.0*cells[p]*V_cell/(4.0*PI))**(1.0/3.0)


        self.mean_overdensity = np.asarray(mean_overdensity, dtype=np.float32)
        self.mean_radius      = np.asarray(mean_radius,      dtype=np.float32)


# This routine creates a density field filled with 0. It then places random
# spheres (not overlapping) with a profile delta(r) = -1*(1-(r/R)^3)
# It returns that density field. A void finder can be run a the identified
# voids can be compared with the input ones.
# For this density profile, the average delta within R is -0.5
@cython.boundscheck(False)
@cython.cdivision(True)
@cython.wraparound(False)
class random_spheres:
    def __init__(self, float BoxSize, float Rmin, float Rmax, int Nvoids, int dims):

        cdef int new_void
        cdef int i, j, l, m, n, Nshells, ii, i1, jj, j1, kk, k1
        cdef float R, diff_x, diff_y, diff_z
        cdef float[:,:,::1] delta
        cdef float[::1] pos, radii
        cdef float [:,::1] positions

        # define the density field
        delta = np.zeros((dims,dims,dims), dtype=np.float32)

        # define the arrays with the positions and radii of the voids
        positions = np.zeros((Nvoids,3), dtype=np.float32)
        radii     = np.zeros(Nvoids,     dtype=np.float32)

        # do a loop over all the voids
        for i in range(Nvoids):

            while(1):

                new_void = 1

                # generate the position and radius of new void
                pos = np.random.random(3).astype(np.float32)*BoxSize
                R   = np.random.random(1).astype(np.float32)*(Rmax-Rmin) + Rmin

                for j in range(i):

                    diff_x = pos[0]-positions[j,0]
                    if diff_x<-BoxSize/2.0:  diff_x += BoxSize
                    if diff_x>BoxSize/2.0:   diff_x -= BoxSize

                    diff_y = pos[1]-positions[j,1]
                    if diff_y<-BoxSize/2.0:  diff_y += BoxSize
                    if diff_y>BoxSize/2.0:   diff_y -= BoxSize

                    diff_z = pos[2]-positions[j,2]
                    if diff_z<-BoxSize/2.0:  diff_z += BoxSize
                    if diff_z>BoxSize/2.0:   diff_z -= BoxSize

                    dist = sqrt(diff_x*diff_x + diff_y*diff_y + diff_z*diff_z)

                    if dist<R+radii[j]:
                        new_void = 0
                        break

                if new_void==1:  
                    positions[i] = pos
                    radii[i]     = R
                    break

            # find the position and radius of the void in grid units
            l       = <int>(positions[i,0]*dims*1.0/BoxSize)
            m       = <int>(positions[i,1]*dims*1.0/BoxSize)
            n       = <int>(positions[i,2]*dims*1.0/BoxSize)
            Nshells = <int>(radii[i]*dims*1.0/BoxSize)

            for ii in range(-Nshells, Nshells+1):
                i1 = (ii+l+dims)%dims
                for jj in range(-Nshells, Nshells+1):
                    j1 = (jj+m+dims)%dims
                    for kk in range(-Nshells, Nshells+1):
                        k1 = (kk+n+dims)%dims

                        dist = sqrt(ii**2 + jj**2 + kk**2)*BoxSize*1.0/dims

                        if dist<R:
                            delta[i1,j1,k1]=-1.0*(1.0-(dist/R)**3)

        self.void_pos    = np.asarray(positions)
        self.void_radius = np.asarray(radii)
        self.delta       = np.asarray(delta)



#model read_FoF"""
#readfof(Base,snapnum,endian=None)
#Read FoF files from Gadget, P-FoF
     #Parameters:
        #basedir: where your FoF folder located
        #snapnum: snapshot number
        #long_ids: whether particles ids are uint32 or uint64
        #swap: False or True
     #return structures:
        #TotNgroups,TotNids,GroupLen,GroupOffset,GroupMass,GroupPos,GroupIDs...
     #Example:
        #--------
        #FoF_halos=readfof("/data1/villa/b500p512nu0.6z99tree",17,long_ids=True,swap=False)
        #Masses=FoF_halos.GroupMass
        #IDs=FoF_halos.GroupIDs
        #--------
        #updated time 19 Oct 2012 by wgcui

#For simulations with SFR, set SFR=True
#The physical velocities of the halos are found multiplying the field 
#GroupVel by (1+z)

import numpy as np
import os, sys
from struct import unpack

class FoF_catalog:
    def __init__(self, basedir, snapnum, long_ids=False, swap=False,
                 SFR=False, read_IDs=True, prefix='/groups_'):

        if long_ids:  format = np.uint64
        else:         format = np.uint32

        exts=('000'+str(snapnum))[-3:]

        #################  READ TAB FILES ################# 
        fnb, skip, Final = 0, 0, False
        dt1 = np.dtype((np.float32,3))
        dt2 = np.dtype((np.float32,6))
        prefix = basedir + prefix + exts + "/group_tab_" + exts + "."
        while not(Final):
            f=open(prefix+str(fnb), 'rb')
            self.Ngroups    = np.fromfile(f, dtype=np.int32,  count=1)[0]
            self.TotNgroups = np.fromfile(f, dtype=np.int32,  count=1)[0]
            self.Nids       = np.fromfile(f, dtype=np.int32,  count=1)[0]
            self.TotNids    = np.fromfile(f, dtype=np.uint64, count=1)[0]
            self.Nfiles     = np.fromfile(f, dtype=np.uint32, count=1)[0]

            TNG, NG = self.TotNgroups, self.Ngroups
            if fnb == 0:
                self.GroupLen    = np.empty(TNG, dtype=np.int32)
                self.GroupOffset = np.empty(TNG, dtype=np.int32)
                self.GroupMass   = np.empty(TNG, dtype=np.float32)
                self.GroupPos    = np.empty(TNG, dtype=dt1)
                self.GroupVel    = np.empty(TNG, dtype=dt1)
                self.GroupTLen   = np.empty(TNG, dtype=dt2)
                self.GroupTMass  = np.empty(TNG, dtype=dt2)
                if SFR:  self.GroupSFR = np.empty(TNG, dtype=np.float32)
                    
            if NG>0:
                locs=slice(skip,skip+NG)
                self.GroupLen[locs]    = np.fromfile(f,dtype=np.int32,count=NG)
                self.GroupOffset[locs] = np.fromfile(f,dtype=np.int32,count=NG)
                self.GroupMass[locs]   = np.fromfile(f,dtype=np.float32,count=NG)
                self.GroupPos[locs]    = np.fromfile(f,dtype=dt1,count=NG)
                self.GroupVel[locs]    = np.fromfile(f,dtype=dt1,count=NG)
                self.GroupTLen[locs]   = np.fromfile(f,dtype=dt2,count=NG)
                self.GroupTMass[locs]  = np.fromfile(f,dtype=dt2,count=NG)
                if SFR:
                    self.GroupSFR[locs]=np.fromfile(f,dtype=np.float32,count=NG)
                skip+=NG

                if swap:
                    self.GroupLen.byteswap(True)
                    self.GroupOffset.byteswap(True)
                    self.GroupMass.byteswap(True)
                    self.GroupPos.byteswap(True)
                    self.GroupVel.byteswap(True)
                    self.GroupTLen.byteswap(True)
                    self.GroupTMass.byteswap(True)
                    if SFR:  self.GroupSFR.byteswap(True)
                        
            curpos = f.tell()
            f.seek(0,os.SEEK_END)
            if curpos != f.tell():
                raise Exception("Warning: finished reading before EOF for tab file",fnb)
            f.close()
            fnb+=1
            if fnb==self.Nfiles: Final=True


        #################  READ IDS FILES ################# 
        if read_IDs:

            fnb,skip=0,0
            Final=False
            while not(Final):
                fname=basedir+"/groups_" + exts +"/group_ids_"+exts +"."+str(fnb)
                f=open(fname,'rb')
                Ngroups     = np.fromfile(f,dtype=np.uint32,count=1)[0]
                TotNgroups  = np.fromfile(f,dtype=np.uint32,count=1)[0]
                Nids        = np.fromfile(f,dtype=np.uint32,count=1)[0]
                TotNids     = np.fromfile(f,dtype=np.uint64,count=1)[0]
                Nfiles      = np.fromfile(f,dtype=np.uint32,count=1)[0]
                Send_offset = np.fromfile(f,dtype=np.uint32,count=1)[0]
                if fnb==0:
                    self.GroupIDs=np.zeros(dtype=format,shape=TotNids)
                if Ngroups>0:
                    if long_ids:
                        IDs=np.fromfile(f,dtype=np.uint64,count=Nids)
                    else:
                        IDs=np.fromfile(f,dtype=np.uint32,count=Nids)
                    if swap:
                        IDs=IDs.byteswap(True)
                    self.GroupIDs[skip:skip+Nids]=IDs[:]
                    skip+=Nids
                curpos = f.tell()
                f.seek(0,os.SEEK_END)
                if curpos != f.tell():
                    raise Exception("Warning: finished reading before EOF for IDs file",fnb)
                f.close()
                fnb+=1
                if fnb==Nfiles: Final=True


# This function is used to write one single file for the FoF instead of having
# many files. This will make faster the reading of the FoF file
def writeFoFCatalog(fc, tabFile, idsFile=None):
    if fc.TotNids > (1<<32)-1: raise Exception('TotNids overflow')

    f = open(tabFile, 'wb')
    np.asarray(fc.TotNgroups).tofile(f)
    np.asarray(fc.TotNgroups).tofile(f)
    np.asarray(fc.TotNids, dtype=np.int32).tofile(f)
    np.asarray(fc.TotNids).tofile(f)
    np.asarray(1, dtype=np.uint32).tofile(f)
    fc.GroupLen.tofile(f)
    fc.GroupOffset.tofile(f)
    fc.GroupMass.tofile(f)
    fc.GroupPos.tofile(f)
    fc.GroupVel.tofile(f)
    fc.GroupTLen.tofile(f)
    fc.GroupTMass.tofile(f)
    if hasattr(fc, 'GroupSFR'):
        fc.GroupSFR.tofile(f)
    f.close()

    if idsFile:
        f = open(idsFile, 'wb')
        np.asarray(fc.TotNgroups).tofile(f)
        np.asarray(fc.TotNgroups).tofile(f)
        np.asarray(fc.TotNids, dtype=np.uint32).tofile(f) 
        np.asarray(fc.TotNids).tofile(f)
        np.asarray(1, dtype=np.uint32).tofile(f)
        np.asarray(0, dtype=np.uint32).tofile(f) 
        fc.GroupIDs.tofile(f)
        f.close()



# This is an example on how to change files
"""
root = '/mnt/xfs1/home/fvillaescusa/data/Neutrino_simulations/Sims_Dec16_2/'
################################## INPUT ######################################
folders = ['0.0eV/','0.06eV/','0.10eV/','0.10eV_degenerate/',
           '0.15eV/','0.6eV/',
           '0.0eV_0.798/','0.0eV_0.807/','0.0eV_0.818/','0.0eV_0.822/',
           '0.0eV_s8c/','0.0eV_s8m/']
###############################################################################

# do a loop over the different cosmologies
for folder in folders:

    # do a loop over the different realizations
    for i in range(1,101):

        snapdir = root + folder + '%d/'%i
        
        # do a loop over the different redshift
        for snapnum in [0,1,2,3]:

            FoF_folder     = snapdir+'groups_%03d'%snapnum
            old_FoF_folder = snapdir+'original_groups_%03d'%snapnum
            if os.path.exists(FoF_folder):
                print('%s\t%d\t%d\texists'%(folder,i,snapnum))

                if os.path.exists(old_FoF_folder):
                    continue

                # create new FoF file
                f_tab = '%s/group_tab_%03d.0'%(snapdir,snapnum)
                f_ids = '%s/group_ids_%03d.0'%(snapdir,snapnum)
                FoF = readfof.FoF_catalog(snapdir,snapnum,long_ids=False,
                                          swap=False,SFR=False)
                writeFoFCatalog(FoF, f_tab, idsFile=f_ids)
           
                # rename FoF folder, create new FoF folder and move files to it
                os.system('mv '+FoF_folder+' '+old_FoF_folder)
                os.system('mkdir '+FoF_folder)
                os.system('mv '+f_tab+' '+f_ids+' '+FoF_folder)
"""
